from __future__ import annotations

from bagua.torch_api.communication import (
    get_backend,
    broadcast,
    _get_default_group,
    _rank_not_in_group,
    BaguaProcessGroup,
)
import bagua
from bagua.torch_api.utils import to_bagua_datatype, StatisticalAverage
from bagua.torch_api.model_parallel.moe import is_moe_param
import gorilla
import torch
import torch.nn
import itertools
from typing import List, Tuple, Optional
from bagua.torch_api.data_parallel.bagua_distributed import BaguaDistributedDataParallel


@gorilla.patches(torch.nn.Module, filter=lambda name, obj: "bagua" in name)
class BaguaModule:
    """
    This class patches `torch.nn.Module <https://pytorch.org/docs/stable/generated/torch.nn.Module.html?highlight=module#torch.nn.Module>`_ with several methods to enable Bagua
    functionalities.

    :ivar bagua_optimizers: The optimizers passed in by :meth:`~bagua.torch_api.distributed.BaguaModule.with_bagua`.
    :vartype bagua_optimizers: List[torch.optim.Optimizer]

    :ivar bagua_algorithm: The algorithm implementation used by the module, reified by the algorithm passed in
        by :meth:`~bagua.torch_api.distributed.BaguaModule.with_bagua`.
    :vartype bagua_algorithm: bagua.torch_api.algorithms.AlgorithmImpl

    :ivar process_group: The process group used by the module.
    :vartype process_group: bagua.torch_api.communication.BaguaProcessGroup

    :ivar bagua_module_name: The module's name. Bagua uses the module name to distinguish different modules.
    :vartype bagua_optimizers: str

    :ivar parameters_to_ignore: The parameter names in ``"{module_name}.{param_name}"`` format to ignore
        when calling ``self.bagua_build_params()``.
    :vartype parameters_to_ignore: List[str]

    :ivar bagua_train_step_counter: Number of iterations in training mode.
    :vartype bagua_train_step_counter: int

    :ivar bagua_buckets: All Bagua buckets in a list.
    :vartype bagua_buckets: List[bagua.torch_api.bucket.BaguaBucket]
    """

    __id_iter = itertools.count()

    def with_bagua(  # pytype: disable=module-attr
        self,
        optimizers: List[torch.optim.Optimizer],
        algorithm: "bagua.torch_api.algorithms.Algorithm",
        process_group: Optional[BaguaProcessGroup] = None,
        do_flatten: bool = True,
    ) -> BaguaModule:
        r"""``with_bagua`` enables easy distributed data parallel training on a
        `torch.nn.Module <https://pytorch.org/docs/stable/generated/torch.nn.Module.html?highlight=module#torch.nn.Module>`_.

        Arguments:
            optimizers: Optimizer(s) used by the
                module. It can contain one or more PyTorch optimizers.
            algorithm: Distributed algorithm
                used to do the actual communication and update.
            process_group: The process group to be used for distributed data all-reduction. If ``None``, the default process group,
                which is created by :func:`bagua.torch_api.init_process_group`, will be used. (default: ``None``)
            do_flatten: Whether to flatten the Bagua buckets. The flatten operation will reset data pointer of bucket
                tensors so that they can use faster code paths. Default: ``True``.

        Returns:
            The original module, with Bagua related environments initialized.

        .. note::
            If we want to ignore some layers for communication, we can first check
            these layer's corresponding keys in the module's ``state_dict`` (they are
            in ``"{module_name}.{param_name}"`` format), then assign the list of
            keys to ``your_module._bagua_params_and_buffers_to_ignore``.

        Examples::

            >>> model = torch.nn.Sequential(
            ...      torch.nn.Linear(D_in, H),
            ...      torch.nn.ReLU(),
            ...      torch.nn.Linear(H, D_out),
            ...    )
            >>> optimizer = torch.optim.SGD(
            ...      model.parameters(),
            ...      lr=0.01,
            ...      momentum=0.9
            ...    )
            >>> model = model.with_bagua(
            ...      [optimizer],
            ...      GradientAllReduce()
            ...    )
        """
        if process_group is None:
            process_group = _get_default_group()

        if not hasattr(self, "_bagua_module_name"):
            self._bagua_module_name = "{}_{}".format(self.__class__.__name__, id(self))

        self.bagua_ddp = BaguaDistributedDataParallel(
            self,
            optimizers=optimizers,
            algorithm=algorithm,
            process_group=process_group,
            bagua_module_name=self.bagua_module_name,
            gradient_as_bucket_view=do_flatten,
        )

        return self

    @torch.jit.unused
    @property
    def bagua_module_name(self):
        """
        The module's name. Bagua uses the module name to distinguish different modules.
        """
        return self._bagua_module_name

    @bagua_module_name.setter
    def bagua_module_name(self, name: str):
        self._bagua_module_name = name

    @torch.jit.unused
    @property
    def bagua_algorithm(self):
        return self.bagua_ddp.bagua_algorithm

    @torch.jit.unused
    @property
    def bagua_optimizers(self):
        return self.bagua_ddp.bagua_optimizers

    @torch.jit.unused
    @property
    def bagua_buckets(self):
        return self.bagua_ddp.bagua_buckets


_base = gorilla._get_base(BaguaModule)
_decorator_data = gorilla.get_decorator_data(_base)
for patch in _decorator_data.patches:
    gorilla.apply(patch)



# pytype: disable=attribute-error
import torch
import time
import io
import pickle
import collections
import logging
from typing import List, Tuple, Optional, Dict
from torch.nn.modules import Module

import bagua
from bagua.torch_api import env
from bagua.torch_api.communication import (
    get_backend,
    get_hyperparameters_service_client,
    broadcast,
    BaguaProcessGroup,
)
from bagua.torch_api.model_parallel.moe import is_moe_param
from bagua.bagua_define import (
    TensorDeclaration,
    BaguaHyperparameter,
)
from bagua.torch_api.utils import to_bagua_datatype, StatisticalAverage


class BaguaDistributedDataParallel:
    def __init__(
        self,
        module: Module,
        optimizers: List[torch.optim.Optimizer],
        algorithm: "bagua.torch_api.algorithms.Algorithm",
        process_group: BaguaProcessGroup,
        bagua_module_name: Optional[str] = None,
        gradient_as_bucket_view: bool = True,
        find_unused_parameters: bool = False,
    ) -> None:
        self.module = module
        self.bagua_module_name = bagua_module_name

        self.bagua_optimizers = optimizers
        self.bagua_algorithm = algorithm.reify(process_group)
        self.process_group = process_group
        self.gradient_as_bucket_view = gradient_as_bucket_view
        self.find_unused_parameters = find_unused_parameters
        self.parameters_to_ignore = (
            []
        )  #: the parameter names to ignore during communication
        if hasattr(self.module, "_bagua_params_and_buffers_to_ignore"):
            self.parameters_to_ignore.extend(
                self.module._bagua_params_and_buffers_to_ignore
            )
        if hasattr(
            self.module, "_ddp_params_and_buffers_to_ignore"
        ):  # for compatibility with PyTorch DDP
            self.parameters_to_ignore.extend(
                self.module._ddp_params_and_buffers_to_ignore
            )

        self.bagua_train_step_counter = 0
        """
        Number of iterations in training mode.
        """
        self.bagua_buckets = []
        """
        All Bagua buckets in a list.
        """
        self._bagua_autotune_last_report_time = time.time()
        self._bagua_autotune_completed = False

        class BaguaDistributedDataParallelStates:
            """Empty class whose instances are used for keeping track of BaguaDistributedDataParallel's internal states."""

            pass

        if hasattr(self.module, "_bagua_states"):
            self._reset_algorithm_state()

        self.module._bagua_states = BaguaDistributedDataParallelStates()
        bagua_states = self.module._bagua_states
        bagua_states._bagua_autograd_hooks = []
        bagua_states._bagua_framework_hooks = []

        self._bagua_backend = get_backend(self.bagua_module_name)
        self._bagua_hyperparameters = BaguaHyperparameter()
        self._speed_metrics_switch_on = env.get_autotune_level() >= 1
        self._speed_metrics = StatisticalAverage()
        self.require_backward_grad_sync = True
        self.autograd_graph_params: Dict[str, torch.nn.Parameter] = {}

        ddp = self

        def autotune_hook(self, input):
            if self.training:
                if env.get_autotune_level() >= 1 and not ddp._bagua_autotune_completed:
                    ddp._bagua_autotune_step()

        def clear_post_backward_callback_queued_hook(self, input):
            ddp._is_post_backward_callback_queued = False

        def num_iteration_step_hook(self, input):
            if self.training:
                ddp.bagua_train_step_counter += 1

        def algorithm_reset_hook(self, input):
            if ddp.bagua_algorithm.need_reset() and self.training:
                ddp._bagua_init_algorithm()

        def algorithm_forward_pre_hook(self, input):
            if self.training:
                ddp.bagua_algorithm.init_forward_pre_hook(ddp)(input)

        def record_speed_metrics_event(self, _):
            if not ddp._speed_metrics_switch_on:
                return

            if hasattr(ddp, "_last_event_pair"):
                (start, stop) = ddp._last_event_pair
                try:
                    elapsed_time_s = start.elapsed_time(stop) / 1000.0
                    total_bytes = sum(bucket.bytes() for bucket in ddp.bagua_buckets)
                    total_gbytes = total_bytes / 1024.0 ** 3
                    speed = total_gbytes / elapsed_time_s
                    ddp._speed_metrics.record(speed)
                except RuntimeError as err:
                    logging.debug("Ignore cuda err={}".format(err))

            start_event = torch.cuda.Event(enable_timing=True)
            ddp._speed_metrics_end_event = torch.cuda.Event(enable_timing=True)
            torch.cuda.current_stream().record_event(start_event)
            ddp._last_event_pair = (start_event, ddp._speed_metrics_end_event)

        def clear_autograd_graph_params(self, _):
            ddp.autograd_graph_params.clear()

        bagua_states._bagua_framework_hooks.extend(
            [
                self.module.register_forward_pre_hook(clear_autograd_graph_params),
                self.module.register_forward_pre_hook(num_iteration_step_hook),
                self.module.register_forward_pre_hook(algorithm_reset_hook),
                self.module.register_forward_pre_hook(algorithm_forward_pre_hook),
                self.module.register_forward_pre_hook(record_speed_metrics_event),
                self.module.register_forward_pre_hook(autotune_hook),
                self.module.register_forward_pre_hook(
                    clear_post_backward_callback_queued_hook
                ),
            ]
        )

        # autotune service
        self._bagua_autotune_client = get_hyperparameters_service_client()

        self._bagua_init_algorithm()

    def bagua_build_params(self) -> List[Tuple[str, torch.nn.Parameter]]:
        """
        Build tuple of ``(parameter_name, parameter)`` for all parameters that
        require grads and not in the ``_bagua_params_and_buffers_to_ignore`` attribute.
        """
        modules_and_parameters = [
            (module, parameter)
            for module_name, module in self.module.named_modules()
            for parameter in [
                (f"{module_name}.{param_name}", param)
                # Note that we access module.named_parameters instead of
                # parameters(module). parameters(module) is only needed in the
                # single-process multi device case, where it accesses replicated
                # parameters through _former_parameters.
                for param_name, param in module.named_parameters(recurse=False)
                if param.requires_grad
                and f"{module_name}.{param_name}" not in self.parameters_to_ignore
                and (not is_moe_param(param))
            ]
        ]

        if self.find_unused_parameters and len(self.autograd_graph_params) != 0:
            modules_and_parameters = filter(
                lambda it: it[1][0] in self.autograd_graph_params,
                modules_and_parameters,
            )

        # Deduplicate any parameters that might be shared across child modules.
        memo = set()
        # "p not in memo" is the deduplication check.
        # "not memo.add(p)" is always True, and it's only there to cause "add(p)" if needed.
        modules_and_parameters = [
            (m, p)
            for m, p in modules_and_parameters
            if p[1] not in memo and not memo.add(p[1])
        ]

        # Build list of parameters.
        parameters = [parameter for _, parameter in modules_and_parameters]

        # Checks if a module will produce a sparse gradient.
        def produces_sparse_gradient(module):
            if isinstance(module, torch.nn.Embedding) or isinstance(
                module, torch.nn.EmbeddingBag
            ):
                return module.sparse
            return False

        # Build list of booleans indicating whether or not to expect sparse
        # gradients for the corresponding parameters.
        expect_sparse_gradient = [
            produces_sparse_gradient(module) for module, _ in modules_and_parameters
        ]

        if any(expect_sparse_gradient):
            raise NotImplementedError("sparse gradient not supported yet")

        return parameters

    # Copyright 2020 Uber Technologies, Inc. All Rights Reserved.
    # Copyright (c) 2021 Kuaishou AI Platform & DS3 Lab.
    #
    # Licensed under the Apache License, Version 2.0 (the "License");
    # you may not use this file except in compliance with the License.
    # You may obtain a copy of the License at
    #
    #     http://www.apache.org/licenses/LICENSE-2.0
    #
    # Unless required by applicable law or agreed to in writing, software
    # distributed under the License is distributed on an "AS IS" BASIS,
    # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    # See the License for the specific language governing permissions and
    # limitations under the License.
    # ==============================================================================
    def _bagua_broadcast_optimizer_state(self, optimizer):
        # L-BFGS cannot be easily supported without serializing
        # the entire state_dict, as its structure is deeply nested and contains
        # None type parameter values.
        if isinstance(optimizer, torch.optim.LBFGS):
            raise ValueError("cannot broadcast torch.optim.LBFGS state")
        optimizer_state_dict = optimizer.state_dict()

        # Initialize newly created optimizers.
        if len(optimizer_state_dict["state"]) == 0:
            for group in optimizer.param_groups:
                for p in group["params"]:
                    if p.requires_grad and id(p) not in optimizer_state_dict["state"]:
                        p.bagua_ensure_grad()
                        if isinstance(optimizer, torch.optim.SparseAdam):
                            p.grad = p.grad.to_sparse()
            optimizer_state_dict = optimizer.state_dict()
        if len(optimizer_state_dict["state"]) == 0:
            return

        def _state_param_callback(param_id, param_name):
            def _assign_state(v):
                optimizer_state_dict["state"][param_id][param_name] = v

            return _assign_state

        def _hyper_param_callback(index, group_key):
            def _assign_hyper(v):
                optimizer.param_groups[index][group_key] = v

            return _assign_hyper

        params = []
        scalars = collections.OrderedDict()
        call_back_param = {}
        repeat_param_count = collections.defaultdict(int)

        # All "sorted()" operations in this module are used to
        # guarteen the scalar's record order samely in differet ranks.
        for index, param_group in enumerate(optimizer_state_dict["param_groups"]):
            for group_key, group_value in sorted(
                param_group.items(), key=lambda item: item[0]
            ):
                # Hyper-parameters like learning rate are scalars, we need to broadcast them separately.
                if group_key != "params":
                    key = "%s_%d" % (group_key, index)
                    scalars[key] = group_value
                    call_back_param[key] = _hyper_param_callback(index, group_key)
            for param_id in sorted(param_group["params"]):
                if param_id not in optimizer_state_dict["state"]:
                    continue
                param_state = optimizer_state_dict["state"][param_id]
                for param_name, inner_state in sorted(
                    param_state.items(), key=lambda item: item[0]
                ):
                    # Some parameter names, e.g., step, may appear more than once, in which
                    # case we ensure they have a unique identifier defined by
                    # their order.
                    repeat_param_count[param_name] += 1
                    key = "%s_%d" % (str(param_name), repeat_param_count[param_name])
                    if isinstance(inner_state, torch.Tensor):
                        params.append((key, inner_state))
                    else:
                        scalars[key] = inner_state
                        call_back_param[key] = _state_param_callback(
                            param_id, param_name
                        )
        for key, param in params:
            broadcast(param, src=0, comm=self.process_group.get_global_communicator())
        scalars = self._bagua_broadcast_scalars(scalars, src=0)
        for key, p in scalars.items():
            call_back_param[key](p)

    def _bagua_broadcast_scalars(self, scalars, src):
        # Serializes and broadcast scalars by converting them to "ByteTensor".
        b = io.BytesIO()
        pickle.dump(scalars, b)
        t = torch.ByteTensor(bytearray(b.getvalue())).cuda()
        broadcast(t, src=0, comm=self.process_group.get_global_communicator())
        if env.get_rank() != src:
            buf = io.BytesIO(t.cpu().numpy().tobytes())
            scalars = pickle.load(buf)

        return scalars

    def _bagua_broadcast_parameters(self):
        """
        Broadcast model and optimizer states.
        """

        module_states = self.bagua_build_params()
        for name, state in module_states:
            broadcast(state, src=0, comm=self.process_group.get_global_communicator())
        for optimizer in self.bagua_optimizers:
            self._bagua_broadcast_optimizer_state(optimizer)

    def _bagua_autotune_step(self):
        CYCLE_STEP = 100
        start_time = time.time()

        if (
            self.bagua_train_step_counter != 0
            and self.bagua_train_step_counter % CYCLE_STEP == 0
        ):
            # get speed metrics
            time_since_last_update = time.time() - self._bagua_autotune_last_report_time
            speed = self._speed_metrics.get(time_since_last_update)

            # report metrics
            rsp = self._bagua_autotune_client.report_metrics(
                model_name=self.bagua_module_name,
                rank=env.get_rank(),
                train_iter=self.bagua_train_step_counter,
                hyperparameters=self._bagua_hyperparameters.dict(),
                speed=speed,
            )
            assert rsp.status_code == 200, "Unexpected rsp={}".format(rsp)

            # update parameters
            self._reset_buckets()
            self._bagua_autotune_last_report_time = time.time()

        logging.debug("autotune overhead=%s", time.time() - start_time)

    def _bagua_autotune_register_tensors(self):
        """
        Register tensors on autotune server, and return first bucketing suggestions
        """
        autotune_tensor_list = [
            TensorDeclaration(
                {
                    "name": tensor.bagua_tensor_name,
                    "num_elements": tensor.numel(),
                    "dtype": to_bagua_datatype(tensor.dtype),
                }
            )
            for tensor in self._bagua_tensors
        ]

        rsp = self._bagua_autotune_client.register_tensors(
            model_name=self.bagua_module_name,
            tensor_list=autotune_tensor_list,
        )
        assert rsp.status_code == 200, "Unexpected rsp={}".format(rsp)

    def _bagua_autotune_get_buckets(self):
        rsp = self._bagua_autotune_client.ask_hyperparameters(
            model_name=self.bagua_module_name,
            rank=env.get_rank(),
            train_iter=self.bagua_train_step_counter,
        )
        assert rsp.status_code == 200, "Unexpected rsp={}".format(rsp)
        recommended_hyperparameters = rsp.json()["recommended_hyperparameters"]
        is_autotune_completed = rsp.json()["is_autotune_completed"]

        self._bagua_hyperparameters.update(recommended_hyperparameters)

        self._bagua_autotune_completed = is_autotune_completed
        recommended_buckets = map(
            lambda x: list(map(lambda y: self._bagua_tensor_map[y["name"]], x)),
            recommended_hyperparameters["buckets"],
        )
        return list(recommended_buckets)

    def _bagua_init_algorithm(self):
        self._bagua_broadcast_parameters()

        self._bagua_tensors = self.bagua_algorithm.init_tensors(self)
        self._bagua_tensor_map = dict(
            [(tensor.bagua_tensor_name, tensor) for tensor in self._bagua_tensors]
        )
        self._bagua_autotune_register_tensors()
        self._reset_buckets()

        self._register_autograd_hooks()
        self._register_optimizer_hooks()

    def _delay_allreduce(self):
        for param_name, parameter in self.bagua_build_params():
            self.bagua_algorithm.init_backward_hook(self)(param_name, parameter)
            self.bagua_algorithm.init_post_backward_hook(self)()

    def _cleanup_autograd_hooks(self):
        bagua_states = self.module._bagua_states
        for hook in bagua_states._bagua_autograd_hooks:
            hook.remove()
        bagua_states._bagua_autograd_hooks.clear()

    def _register_autograd_hooks(self):
        bagua_states = self.module._bagua_states
        self._cleanup_autograd_hooks()

        for name, param in self.module.named_parameters():

            def real_hook_factory(param_name, parameter):
                def real_hook(*unused):
                    if not self.require_backward_grad_sync:
                        return

                    if self.find_unused_parameters:
                        self.autograd_graph_params[param_name] = parameter

                    self.bagua_algorithm.init_backward_hook(self)(param_name, parameter)

                    def real_post_backward_hook(*unused):
                        self.bagua_algorithm.init_post_backward_hook(self)()
                        if self._speed_metrics_switch_on:
                            torch.cuda.current_stream().record_event(
                                self._speed_metrics_end_event
                            )

                        if self.find_unused_parameters:
                            if (
                                set(self.autograd_graph_params.keys())
                                != self.params_in_use
                            ):
                                self._reset_buckets()
                                self._delay_allreduce()

                    if not self._is_post_backward_callback_queued:
                        torch.autograd.Variable._execution_engine.queue_callback(
                            real_post_backward_hook
                        )
                        self._is_post_backward_callback_queued = True

                return real_hook

            if param.requires_grad:
                param_tmp = param.expand_as(param)
                grad_acc = param_tmp.grad_fn.next_functions[0][0]
                hook = grad_acc.register_hook(real_hook_factory(name, param))
                hook.grad_acc = grad_acc
                bagua_states._bagua_autograd_hooks.append(hook)

    def _register_optimizer_hooks(self):
        optimizer_hook = self.bagua_algorithm.init_post_optimizer_step_hook(self)

        from types import MethodType

        for optimizer in self.bagua_optimizers:
            if not hasattr(optimizer, "_bagua_original_step"):
                optimizer._bagua_original_step = optimizer.step

            def new_step_factory(optimizer):
                def new_step(self, *args, **kwargs):
                    result = self._bagua_original_step(*args, **kwargs)

                    optimizer_hook(self)
                    return result

                return MethodType(new_step, optimizer)

            optimizer.step = new_step_factory(optimizer)

    def _reset_buckets(self):
        raw_buckets = self._bagua_autotune_get_buckets()
        self.bagua_buckets = self.bagua_algorithm.tensors_to_buckets(
            raw_buckets, self.gradient_as_bucket_view
        )
        for bucket in self.bagua_buckets:
            self.bagua_algorithm.init_operations(
                self,
                bucket,
            )
        self._bagua_backend.register_ordered_buckets(
            [bucket.backend_bucket for bucket in self.bagua_buckets]
        )
        self.params_in_use = set([name for name, _ in self.bagua_build_params()])

    def _reset_algorithm_state(self):
        bagua_states = self.module._bagua_states
        if hasattr(bagua_states, "_bagua_framework_hooks"):
            for hook in bagua_states._bagua_framework_hooks:
                hook.remove()

        if hasattr(bagua_states, "_bagua_autograd_hooks"):
            self._cleanup_autograd_hooks()



# pytype: disable=module-attr
import torch
from torch.nn.modules import Module
from torch.distributed import ProcessGroup as TorchProcessGroup
from torch.nn.parallel import DistributedDataParallel as TorchDistributedDataParallel
from typing import Callable, List, Optional, Union
import warnings

import bagua
from bagua.torch_api.algorithms.gradient_allreduce import GradientAllReduceAlgorithm
from contextlib import contextmanager
from bagua.torch_api.communication import (
    _get_default_group,
    BaguaProcessGroup,
)
from .bagua_distributed import BaguaDistributedDataParallel


class DistributedDataParallel_V1_9_0_Interface(Module):
    r"""
    `PyTorch v1.9.0 DDP <https://github.com/pytorch/pytorch/blob/v1.9.0/torch/nn/parallel/distributed.py#L125>`_ compatible interface.
    """

    def __init__(self) -> None:
        super(DistributedDataParallel_V1_9_0_Interface, self).__init__()

    @contextmanager
    def no_sync(self):
        raise NotImplementedError

    def forward(self, *inputs, **kwargs):
        raise NotImplementedError

    def scatter(self, inputs, kwargs, device_ids):
        raise NotImplementedError

    def to_kwargs(self, inputs, kwargs, device_id):
        raise NotImplementedError

    def gather(self, outputs, output_device):
        raise NotImplementedError

    def train(self, mode=True):
        super(DistributedDataParallel_V1_9_0_Interface, self).train(mode)
        return self

    @contextmanager
    def join(
        self,
        divide_by_initial_world_size=True,
        enable=True,
        throw_on_early_termination=False,
    ):
        raise NotImplementedError

    def register_comm_hook(self, state: object, hook: Callable):
        raise NotImplementedError

    def will_sync_module_buffers(self):
        raise NotImplementedError


def to_bagua_process_group(
    process_group: Union[TorchProcessGroup, BaguaProcessGroup, None] = None
):
    """Convert a PyTorch process group to a Bagua process group.

    Args:
        process_group (Union[TorchProcessGroup, BaguaProcessGroup, None], optional): PyTorch
            process group or Bagua process group. The default PyTorch process group is used if ``None`` is passed in.

    Raises:
        Exception: raise unexpect input exception if input is not
            ``TorchProcessGroup``, ``BaguaProcessGroup`` or ``None``.

    Returns:
        BaguaProcessGroup: process group for communication in bagua.
    """

    if process_group is None:
        return _get_default_group()
    elif type(process_group) in [
        TorchProcessGroup,
        torch.distributed.ProcessGroupNCCL,
    ]:
        return process_group.bagua_patch().bagua_pg  # pytype: disable=attribute-error
    elif type(process_group) is BaguaProcessGroup:
        return process_group
    else:
        raise Exception("unexpect input {}".format(type(process_group)))


class DistributedDataParallel_V1_9_0(DistributedDataParallel_V1_9_0_Interface):
    r"""
    `PyTorch v1.9.0 DDP <https://github.com/pytorch/pytorch/blob/v1.9.0/torch/nn/parallel/distributed.py#L125>`_ compatible interface.
    """

    def __init__(
        self,
        module,
        device_ids=None,
        output_device=None,
        dim=0,
        broadcast_buffers=True,
        process_group=None,
        bucket_cap_mb=25,
        find_unused_parameters=False,
        check_reduction=False,
        gradient_as_bucket_view=True,
        # The following bagua parameters
        optimizers: List[torch.optim.Optimizer] = [],
        algorithm: "bagua.torch_api.algorithms.Algorithm" = GradientAllReduceAlgorithm(),
    ) -> None:
        """Bagua internal use function. Construction use :class:`DistributedDataParallel`."""
        super(DistributedDataParallel_V1_9_0, self).__init__()
        assert any((p.requires_grad for p in module.parameters())), (
            "DistributedDataParallel is not needed when a module "
            "doesn't have any parameter that requires a gradient."
        )

        if device_ids is not None and len(device_ids) > 1:
            raise ValueError("device_ids can only be None or contain a single element.")

        self.is_multi_device_module = len({p.device for p in module.parameters()}) > 1
        distinct_device_types = {p.device.type for p in module.parameters()}
        if len(distinct_device_types) != 1:
            raise ValueError(
                "DistributedDataParallel's input module must be on "
                "the same type of devices, but input module parameters locate in {}.".format(
                    distinct_device_types
                )
            )
        self.device_type = list(distinct_device_types)[0]

        self.static_graph = False
        self.dim = dim
        self.module = module
        self.device = list(self.module.parameters())[0].device
        assert broadcast_buffers is True, "Not yet supported"
        self.broadcast_buffers = broadcast_buffers
        self.find_unused_parameters = find_unused_parameters

        if not hasattr(module, "_bagua_module_name"):
            module._bagua_module_name = "{}_{}".format(
                self.__class__.__name__, id(module)
            )

        self.inner = BaguaDistributedDataParallel(
            self.module,
            optimizers,
            algorithm,
            process_group=to_bagua_process_group(process_group),
            gradient_as_bucket_view=gradient_as_bucket_view,
            find_unused_parameters=find_unused_parameters,
            bagua_module_name=module.bagua_module_name,
        )

    @property
    def require_backward_grad_sync(self):
        """
        DDP gradient synchronizations switch, see :meth:`no_sync` for usage.
        """
        return self.inner.require_backward_grad_sync

    @property
    def parameters_to_ignore(self):
        """Parameters that will be ignored in DDP."""
        return self.inner.parameters_to_ignore

    def forward(self, *inputs, **kwargs):
        output = self.module(*inputs, **kwargs)
        return output

    @contextmanager
    def no_sync(self):
        r"""
        A context manager to disable gradient synchronizations across DDP
        processes. Within this context, gradients will be accumulated on module
        variables, which will later be synchronized in the first
        forward-backward pass exiting the context.

        Example::

            >>> ddp = bagua.torch_api.data_parallel.DistributedDataParallel(model, pg)
            >>> with ddp.no_sync():
            >>>   for input in inputs:
            >>>     ddp(input).backward()  # no synchronization, accumulate grads
            >>> ddp(another_input).backward()  # synchronize grads
        """
        old_require_backward_grad_sync = self.require_backward_grad_sync
        self.inner.require_backward_grad_sync = False
        try:
            yield
        finally:
            self.inner.require_backward_grad_sync = old_require_backward_grad_sync

    @property
    def bagua_algorithm(self):
        """
        The algorithm implementation used by the module,
        reified by the algorithm passed in from the constructor.
        """
        return self.inner.bagua_algorithm

    @property
    def bagua_optimizers(self):
        """
        Optimizer(s) used by the module. It can contain one or more PyTorch optimizers.
        """
        return self.inner.bagua_optimizers

    @property
    def bagua_buckets(self):
        """
        All Bagua buckets in a list.
        """
        return self.inner.bagua_buckets


def DistributedDataParallel(
    module: Module,
    device_ids: Optional[List[Union[int, torch.device]]] = None,
    output_device: Union[int, torch.device] = None,
    dim: int = 0,
    broadcast_buffers: bool = True,
    process_group: Union[None, TorchProcessGroup] = None,
    bucket_cap_mb: int = 25,
    find_unused_parameters: bool = False,
    check_reduction: bool = False,
    gradient_as_bucket_view: bool = True,
    # The followings are parameters for Bagua
    optimizers: List[torch.optim.Optimizer] = [],
    algorithm: "bagua.torch_api.algorithms.Algorithm" = GradientAllReduceAlgorithm(),
) -> Union[TorchDistributedDataParallel, DistributedDataParallel_V1_9_0]:
    r"""
    This function provides a `PyTorch DDP <https://github.com/pytorch/pytorch/blob/v1.9.0/torch/nn/parallel/distributed.py#L125>`_ compatible
    interface plus several Bagua specific parameters.

    Args:
        module (Module): module to be parallelized
        device_ids (Optional[List[Union[int, torch.device]]], optional): CUDA devices.

                   1) For single-device modules, ``device_ids`` can
                   contain exactly one device id, which represents the only
                   CUDA device where the input module corresponding to this process resides.
                   Alternatively, ``device_ids`` can also be ``None``.

                   2) For multi-device modules and CPU modules,
                   ``device_ids`` must be ``None``.

                   When ``device_ids`` is ``None`` for both cases,
                   both the input data for the forward pass and the actual module
                   must be placed on the correct device.
                   (default: ``None``)
        output_device (Union[int, torch.device], optional): Device location of output for
                      single-device CUDA modules. For multi-device modules and
                      CPU modules, it must be ``None``, and the module itself
                      dictates the output location. (default: ``device_ids[0]``
                      for single-device modules)
        dim (int, optional): Flag that enables syncing (broadcasting)
                          buffers of the module at beginning of the ``forward``
                          function. (default: ``True``)
        broadcast_buffers (bool, optional): Flag that enables syncing (broadcasting)
                          buffers of the module at beginning of the ``forward``
                          function. (default: ``True``)
        process_group (Union[None, TorchProcessGroup], optional): The process group to be used for distributed data
                       all-reduction. If ``None``, the default process group, which
                       is created by :func:`torch.distributed.init_process_group`,
                       will be used. (default: ``None``)
        bucket_cap_mb (int, optional): ``DistributedDataParallel`` will bucket parameters into
                       multiple buckets so that gradient reduction of each
                       bucket can potentially overlap with backward computation.
                       :attr:`bucket_cap_mb` controls the bucket size in
                       MegaBytes (MB). (default: 25)
        find_unused_parameters (bool, optional): Traverse the autograd graph from all
                               tensors contained in the return value of the
                               wrapped module's ``forward`` function. Parameters
                               that don't receive gradients as part of this
                               graph are preemptively marked as being ready to
                               be reduced. In addition, parameters that may have
                               been used in the wrapped module's ``forward``
                               function but were not part of loss computation and
                               thus would also not receive gradients are
                               preemptively marked as ready to be reduced.
                               (default: ``False``)
        check_reduction (bool, optional): This argument is deprecated.
        gradient_as_bucket_view (bool, optional): When set to ``True``, gradients will be views
                      pointing to different offsets of ``allreduce`` communication
                      buckets. This can reduce peak memory usage, where the
                      saved memory size will be equal to the total gradients
                      size. Moreover, it avoids the overhead of copying between
                      gradients and ``allreduce`` communication buckets. When
                      gradients are views, ``detach_()`` cannot be called on the
                      gradients. If hitting such errors, please fix it by
                      referring to the :meth:`~torch.optim.Optimizer.zero_grad`
                      function in ``torch/optim/optimizer.py`` as a solution.
        optimizers (List[torch.optim.Optimizer], optional): Optimizer(s) used by the module. It can contain one or more PyTorch optimizers. Defaults to ``[]``.
        algorithm (bagua.torch_api.algorithms.Algorithm, optional): Data
                parallel distributed algorithm, decide how to communication mode
                and the way the model is updated. Defaults to :class:`~bagua.torch_api.algorithms.gradient_allreduce.GradientAllReduceAlgorithm`.

    Returns:
        Union[TorchDistributedDataParallel, DistributedDataParallel_V1_9_0]: Bagua distributed data parallel instance used for distributed training.

    Example::

        >>> bagua.init_process_group()
        >>> net = bagua.data_parallel.DistributedDataParallel(model)

    Example using faster algorithms in Bagua::

        >>> from bagua.torch_api.algorithms import bytegrad
        >>> bagua.init_process_group()
        >>> net = bagua.data_parallel.DistributedDataParallel(model, algorithm=bytegrad.ByteGradAlgorithm())
        >>> # For more possible algorithms, see https://tutorials.baguasys.com/algorithms/.

    """

    check_list = [
        device_ids is None,
        output_device is None,
        dim == 0,
        broadcast_buffers is True,
        check_reduction is False,
    ]
    if not all(check_list):
        warnings.warn(
            "Some parameters passed into BaguaDistributedDataParallel"
            " have not been supported yet. Bagua has automatically "
            "fallback to upstream PyTorch DistributedDataParallel "
            "implementation. If this is unexpected, please submit "
            "an issue to https://github.com/BaguaSys/bagua. Thanks."
        )
        return TorchDistributedDataParallel(
            module=module,
            device_ids=device_ids,
            output_device=output_device,
            dim=dim,
            broadcast_buffers=broadcast_buffers,
            process_group=process_group,
            bucket_cap_mb=bucket_cap_mb,
            find_unused_parameters=find_unused_parameters,
            check_reduction=check_reduction,
            gradient_as_bucket_view=gradient_as_bucket_view,
        )

    return DistributedDataParallel_V1_9_0(
        module=module,
        device_ids=device_ids,
        output_device=output_device,
        dim=dim,
        broadcast_buffers=broadcast_buffers,
        process_group=process_group,
        bucket_cap_mb=bucket_cap_mb,
        find_unused_parameters=find_unused_parameters,
        check_reduction=check_reduction,
        gradient_as_bucket_view=gradient_as_bucket_view,
        optimizers=optimizers,
        algorithm=algorithm,
    )



import torch
import unittest


def skip_if_cuda_available():
    if torch.cuda.is_available():
        return unittest.skip("skip when cuda is available")

    return lambda func: func


def skip_if_cuda_not_available():
    if not torch.cuda.is_available():
        return unittest.skip("skip when cuda is not available")

    return lambda func: func



import os
import platform
import shutil
import tempfile
import urllib.request
import pathlib
from tqdm import tqdm


_nccl_records = []
library_records = {}


class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    with DownloadProgressBar(
        unit="B", unit_scale=True, miniters=1, desc=url.split("/")[-1]
    ) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


def _make_nccl_url(public_version, filename):
    # https://developer.download.nvidia.com/compute/redist/nccl/v2.8/nccl_2.8.4-1+cuda11.2_x86_64.txz
    return (
        "https://developer.download.nvidia.com/compute/redist/nccl/"
        + "v{}/{}".format(public_version, filename)
    )


def _make_nccl_record(cuda_version, full_version, public_version, filename_linux):
    return {
        "cuda": cuda_version,
        "nccl": full_version,
        "assets": {
            "Linux": {
                "url": _make_nccl_url(public_version, filename_linux),
                "filename": "libnccl.so.{}".format(full_version),
            },
        },
    }


_nccl_records.append(
    _make_nccl_record("11.6", "2.11.4", "2.11", "nccl_2.11.4-1+cuda11.4_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("11.5", "2.11.4", "2.11", "nccl_2.11.4-1+cuda11.4_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("11.4", "2.11.4", "2.11", "nccl_2.11.4-1+cuda11.4_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("11.3", "2.10.3", "2.10", "nccl_2.10.3-1+cuda11.0_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("11.2", "2.10.3", "2.10", "nccl_2.10.3-1+cuda11.0_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("11.1", "2.10.3", "2.10", "nccl_2.10.3-1+cuda11.0_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("11.0", "2.11.4", "2.11", "nccl_2.11.4-1+cuda11.0_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("10.2", "2.11.4", "2.11", "nccl_2.11.4-1+cuda10.2_x86_64.txz")
)
_nccl_records.append(
    _make_nccl_record("10.1", "2.10.3", "2.10", "nccl_2.10.3-1+cuda10.2_x86_64.txz")
)
library_records["nccl"] = _nccl_records


def install_lib(cuda, prefix, library):
    record = None
    lib_records = library_records
    for record in lib_records[library]:
        if record["cuda"] == cuda:
            break
    else:
        raise RuntimeError(
            """
The CUDA version({}) specified is not supported.
Should be one of {}.""".format(
                cuda, str([x["cuda"] for x in lib_records[library]])
            )
        )

    destination = os.path.join(prefix, "nccl")

    if os.path.exists(destination):
        print("The destination directory {} already exists.".format(destination))
        shutil.rmtree(destination)

    target_platform = platform.system()
    asset = record["assets"].get(target_platform, None)
    if asset is None:
        raise RuntimeError(
            """
The current platform ({}) is not supported.""".format(
                target_platform
            )
        )

    print(
        "Installing {} {} for CUDA {} to: {}".format(
            library, record[library], record["cuda"], destination
        )
    )

    url = asset["url"]
    print("Downloading {}...".format(url))
    with tempfile.TemporaryDirectory() as tmpdir:
        filename = os.path.join(tmpdir, os.path.basename(url))
        download_url(url, filename)
        print("Extracting...")
        outdir = os.path.join(tmpdir, "extract")
        shutil.unpack_archive(filename, outdir)
        print("Installing...")
        if library == "nccl":
            subdir = os.listdir(outdir)
            assert len(subdir) == 1
            shutil.move(os.path.join(outdir, subdir[0]), destination)
        else:
            assert False
        print("Cleaning up...")
    print("Done!")


def install_deps():
    nvcc_version = (
        os.popen(
            "nvcc --version | grep release | sed 's/.*release //' |  sed 's/,.*//'"
        )
        .read()
        .strip()
    )
    print("nvcc_version: ", nvcc_version)
    install_lib(
        nvcc_version,
        os.path.join(pathlib.Path.home(), ".local", "share", "bagua"),
        "nccl",
    )


if __name__ == "__main__":
    install_deps()



import torch
from typing import List, Dict, Optional, Any
import copy
import logging
import itertools
from bagua.torch_api.utils import check_contiguous, get_flattened_tensor
from collections import defaultdict
import gorilla


__all__ = ["fuse_optimizer", "fuse_step", "is_fused_optimizer"]


def flatten_tensors_with_closure(tensors, getter_closure, setter_closure):
    if len(tensors) == 0:
        return

    eff_tensors = [getter_closure(t) for t in tensors]
    if check_contiguous(eff_tensors):
        return

    flatten_tensor = get_flattened_tensor(eff_tensors)
    flatten_storage = flatten_tensor.storage()

    offset = 0
    for tensor in tensors:
        with torch.no_grad():
            z = torch.zeros_like(getter_closure(tensor))
            z.set_(flatten_storage, offset, z.shape)
            z._bagua_flattened_tensor = flatten_tensor
            setter_closure(tensor, z)

        offset += z.numel()
        logging.debug(f"flatten done {offset}, dtype: {z.dtype}")

    check_contiguous([getter_closure(t) for t in tensors])


def flatten_params_and_states(optimizer: torch.optim.Optimizer):
    """
    Flatten parameter tensors in the same group into contiguous ones.
    """

    for group in optimizer.param_groups:
        type_params = {}
        # flatten by param group
        for param in group["params"]:

            params_of_type = type_params.get(param.type(), [])
            params_of_type.append(param)
            type_params[param.type()] = params_of_type

        for param_type, params in type_params.items():
            grads = [p.bagua_ensure_grad().grad for p in params]
            state_tensors, state_scalars = get_optimizer_param_states(optimizer, params)

            if state_tensors is None:
                continue

            flatten_tensors_with_closure(
                params,
                getter_closure=lambda p: p.data,
                setter_closure=lambda p, new_data: setattr(p, "data", new_data),
            )

            flatten_tensors_with_closure(
                params,
                getter_closure=lambda p: p.grad.data,
                setter_closure=lambda p, new_grad: setattr(p.grad, "data", new_grad),
            )

            for name, tensors in state_tensors.items():
                flatten_tensors_with_closure(
                    params,
                    getter_closure=lambda p: optimizer.state[p][name].data,
                    setter_closure=lambda p, new_state: setattr(
                        optimizer.state[p][name], "data", new_state
                    ),
                )

            torch.cuda.empty_cache()


def _is_contiguous_tensor(a: torch.Tensor, b: torch.Tensor):
    """
    Checking if tensor :attr:`a` and tensor :attr:`b` are contiguous.
    """
    size_a = a.numel() * a.element_size()
    size_b = b.numel() * b.element_size()

    return (a.data_ptr() == b.data_ptr() + size_b) or (
        b.data_ptr() == a.data_ptr() + size_a
    )


def _find_continuous_tensors(tensors: List[torch.Tensor]):
    tensor_list = zip(tensors, list(range(len(tensors))))
    sorted_tensor_list = sorted(tensor_list, key=lambda x: x[0].data_ptr())

    grouped_indices = []
    tmp_tensors = []
    tmp_indices = []

    for tensor, idx in sorted_tensor_list:
        if len(tmp_tensors) > 0 and not _is_contiguous_tensor(tensor, tmp_tensors[-1]):
            if len(tmp_tensors) > 1:
                grouped_indices.append(tmp_indices)
            tmp_tensors = []
            tmp_indices = []

        tmp_tensors.append(tensor)
        tmp_indices.append(idx)

    if len(tmp_tensors) > 1:
        grouped_indices.append(tmp_indices)

    return grouped_indices


def calculate_mutual_groups(tensors_list: List[List[torch.Tensor]]):
    constraints = []

    size = len(tensors_list[0])
    for tensors in tensors_list:
        assert size == len(
            tensors
        ), "Tensors to calculate mutual groups must have equal size."

        grouped_indices = _find_continuous_tensors(tensors)
        constraints.append(grouped_indices)

    if len(constraints) == 0:
        return constraints

    grouped_indices = constraints[0]
    for i in range(1, len(constraints)):
        grouped_indices = _intersect(grouped_indices, constraints[i])

    return grouped_indices


def _intersect(a: List[List[int]], b: List[List[int]]):
    c = [value for value in a if value in b]
    return c


def group_tensors(tensors: List[torch.Tensor], indices: List[int]):
    if len(indices) == 0:
        return

    to_group = [tensors[idx] for idx in indices]
    assert check_contiguous(to_group), "tensors grouped must be contiguous"

    total_size = sum([t.numel() for t in to_group])
    return to_group[0].storage(), total_size, to_group[0].dtype, to_group[0].device


def _create_tensor(storage, numel, dtype, device):
    logging.debug(f"Create new tensor with numel: {numel}, dtype: {dtype}")
    with torch.no_grad():
        tensor_view = torch.zeros(numel, dtype=dtype, device=device)
        tensor_view.set_(storage, 0, tensor_view.shape)
        return tensor_view


def _can_reuse_tensor(tensor, storage, numel, dtype, device):
    if tensor is None:
        return False

    return (
        tensor.storage().data_ptr() == storage.data_ptr()
        and tensor.numel() == numel
        and tensor.dtype == dtype
        and tensor.device == device
    )


def infer_state_tensors(optimizer, fused_param, original_params, name):
    fused_param_state = get_tensor_state(optimizer, fused_param, name)

    ungrouped = []
    offset = 0
    for p in original_params:
        if name in optimizer.state[p]:
            temp_tensor = get_tensor_state(optimizer, p, name)
        else:
            # use parameter meta to make a guess
            temp_tensor = p

        if temp_tensor.dtype != fused_param_state.dtype:
            logging.warning(
                "Fused optimizer failed to recover original parameter state, due to ambiguous parameter state datatype."
            )
            return

        z = fused_param_state.narrow(0, offset, temp_tensor.numel())
        z = z.view(temp_tensor.shape)
        offset += temp_tensor.numel()
        ungrouped.append(z)

    if offset != fused_param_state.numel():
        logging.warning(
            "Fused optimizer failed to recover original parameter state, due to ambiguous parameter state size."
        )
        return

    return ungrouped


def fuse_optimizer(
    optimizer: torch.optim.Optimizer,
    do_flatten: bool = True,
    check_flatten: bool = True,
):
    """
    Convert any optimizer into a fused optimizer.

    A fused optimizer can fuse multiple parameter updates into one or a few updates. To achieve this, users need to:

    | 1) flatten multiple parameters in the same group into fused parameter by setting :attr:`do_flatten=True`,
         which is also the default behavior of a fused optimizer;
    | 2) perform a fused parameter update by calling :meth:`fuse_step`.

    This fused optimizer is implemented for general use. It can be used used in conjunction with
    a :class:`~bagua.torch_api.distributed.BaguaModule` as well as a
    `torch.nn.parallel.DistributedDataParallel <https://pytorch.org/docs/stable/generated/torch.nn.parallel.DistributedDataParallel.html?highlight=distributeddataparallel#torch.nn.parallel.DistributedDataParallel>`_
    wrapped module, or some other cases (not listed here).

    Args:
        optimizer (torch.optim.Optimizer): Any PyTorch optimizer.
        do_flatten (bool): Whether to flatten the parameters. The flatten operation will reset data pointers of
            parameter tensors so that they can be fused together. Default: ``True``.
        check_flatten (bool): When setting to ``True``, it enables fused optimizer to automatically check if
            parameter tensors are contiguous as they are flattened to. Can only work with :attr:`do_flatten=True`.
            Default: ``True``.

    Returns:
        A Fused optimizer.

    Example::
        >>> optimizer = torch.optim.Adadelta(model.parameters(), ....)
        >>> optimizer = bagua.torch_api.contrib.fuse_optimizer(optimizer, do_flatten=True)
        >>>
        >>> optimizer.fuse_step()

        When use in conjunction with a :class:`~bagua.torch_api.distributed.BaguaModule`, set :attr:`do_flatten=False`
        in :meth:`~bagua.torch_api.distributed.BaguaModule.with_bagua` explicitly:

        >>> optimizer = bagua.torch_api.contrib.fuse_optimizer(optimizer, do_flatten=True)
        >>> model = model.with_bagua([optimizer], GradientAllReduceAlgorithm(), do_flatten=False)
        >>>
        >>> optimizer.fuse_step()

    .. note::
        This function and :meth:`~bagua.torch_api.distributed.BaguaModule.with_bagua` method both will reset data
        pointers of module parameters by default. In order to perform a more effective fused parameter update,
        users need to disable bucket flattening in :meth:`~bagua.torch_api.distributed.BaguaModule.with_bagua`
        by setting its :attr:`do_flatten` to ``False``.

    .. note::
        A fuse optimizer does not change the original behaviors of :attr:`optimizer`, but enabling it to perform a
        fused parameter update through :meth:`fuse_step`. Users can still perform a normal parameter update through
        :meth:`step`.
    """

    if is_fused_optimizer(optimizer):
        raise RuntimeError("trying to fuse an optimizer twice!")

    optimizer._bagua_check_flatten = do_flatten and check_flatten
    optimizer._bagua_fused_count = 0
    optimizer._bagua_cloned_attrs = {}
    optimizer._bagua_fused_optimizer = make_optimizer_instance(optimizer)

    if do_flatten:
        flatten_params_and_states(optimizer)

    if not hasattr(optimizer, "fuse_step"):
        patch = gorilla.Patch(optimizer.__class__, "fuse_step", fuse_step)
        gorilla.apply(patch)

    return optimizer


def is_fused_optimizer(optimizer: torch.optim.Optimizer):
    """
    Checking if :attr:`optimizer` is a fused optimizer or not.
    """
    return hasattr(optimizer, "_bagua_fused_optimizer")


def make_optimizer_instance(optimizer: torch.optim.Optimizer):
    ignore_attrs = [
        "_bagua_check_flatten",
        "_bagua_fused_count",
        "_bagua_cloned_attrs",
    ]
    new_optimizer = copy.copy(optimizer)

    for attr in dir(optimizer):
        if attr not in ignore_attrs and attr not in dir(new_optimizer):
            logging.warning(
                f"Clone attribute {attr} to fused optimizer, should not modify it in `optimizer.step()`."
            )
            setattr(new_optimizer, attr, getattr(optimizer, attr))
            optimizer._bagua_cloned_attrs[attr] = getattr(optimizer, attr)

    # Note: fused optimizer has its own copy of param groups and param state
    new_optimizer.param_groups = []
    for group in optimizer.param_groups:
        new_group = {"params": list(group["params"])}
        new_optimizer.add_param_group(new_group)

    new_optimizer.state = defaultdict(dict)

    return new_optimizer


def fuse_step(optimizer: torch.optim.Optimizer, closure=None):
    r"""Perform a fused parameter update.

    This operation will fuse multiple contiguous parameters into a fused parameter, by creating a tensor
    view sharing the same underlying storage with them, and then perform parameter update on fused parameters.
    If none of the parameter tensors are contiguous, this operation is equivalent to :meth:`step`.

    Args:
        optimizer: A fused optimizer.
        closure (Callable): A closure that reevaluates the model and
            returns the loss. Optional for most optimizers.

    .. note::
        This function will not modify the storage of parameter tensors.
    """

    assert is_fused_optimizer(
        optimizer
    ), "Should init fused optimizer by calling `fuse_optimizer`."

    do_fuse(optimizer)
    optimizer._bagua_fused_optimizer.step(closure)
    check_optimizer(optimizer)
    sync_optimizer_state(optimizer)


def do_fuse(optimizer: torch.optim.Optimizer):
    _fused_optimizer = optimizer._bagua_fused_optimizer

    for group, fused_group in zip(
        optimizer.param_groups, _fused_optimizer.param_groups
    ):
        sync_param_group_scalars(src_group=group, dst_group=fused_group)

        # Find params to fuse
        params = group["params"]
        weights = [p.data for p in params]
        grads = [p.grad for p in params]

        # Find original param state
        state_tensors, state_scalars = get_optimizer_param_states(optimizer, params)

        if state_tensors is None:
            continue

        check_flatten = optimizer._bagua_check_flatten
        if check_flatten and not check_contiguous(weights):
            logging.warning(
                "Parameter weights storage changed after flattened in fused optimizer, may degrade performance."
            )
            check_flatten = False

        if check_flatten and not check_contiguous(grads):
            logging.warning(
                "Parameter gradients storage changed after flattened in fused optimizer, may degrade performance."
            )
            check_flatten = False

        for name, tensors in state_tensors.items():
            if check_flatten and not check_contiguous(tensors):
                logging.warning(
                    "Parameter state {} storage changed after flattened in fused optimizer, may degrade performance.".format(
                        name
                    )
                )
                check_flatten = False

        grouped_indices = calculate_mutual_groups(
            [weights, grads] + list(state_tensors.values())
        )

        if len(grouped_indices) > 0:
            optimizer._bagua_fused_count += 1

        new_params = []
        fused_param_storages = {
            fp.storage().data_ptr(): fp
            for fp in fused_group["params"]
            if hasattr(fp, "_bagua_fused_param_ids")
        }

        active_param_ids = []
        for indices in grouped_indices:
            logging.debug(f"fuse params: {indices}")
            grouped_weight = group_tensors(weights, indices)
            grouped_grad = group_tensors(grads, indices)

            # create fused parameter
            with torch.no_grad():

                if _can_reuse_tensor(
                    fused_param_storages.get(grouped_weight[0].data_ptr()),
                    *grouped_weight,
                ):
                    fp = fused_param_storages[grouped_weight[0].data_ptr()]
                else:
                    fp = torch.nn.Parameter(
                        _create_tensor(*grouped_weight), requires_grad=False
                    )

                if not _can_reuse_tensor(fp.grad, *grouped_grad):
                    fp.grad = _create_tensor(*grouped_grad)

            fp._bagua_fused_param_ids = indices
            new_params.append(fp)
            active_param_ids.append(id(fp))

            # sync state tensors for fused optimizer
            for name, tensors in state_tensors.items():
                grouped_state = group_tensors(tensors, indices)

                if fp not in _fused_optimizer.state or not _can_reuse_tensor(
                    _fused_optimizer.state[fp].get(name, None), *grouped_state
                ):
                    _fused_optimizer.state[fp][name] = _create_tensor(*grouped_state)

            # sync state scalars for fused optimizer
            for name, scalar in state_scalars.items():
                _fused_optimizer.state[fp][name] = scalar

        # add non-contiguous params
        grouped_indices_flat = list(itertools.chain.from_iterable(grouped_indices))
        for idx, param in enumerate(params):
            if idx not in grouped_indices_flat:
                new_params.append(param)
                active_param_ids.append(id(param))

                for name, v in optimizer.state[param].items():
                    _fused_optimizer.state[param][name] = v

        # clear outdated states
        for fp in fused_group["params"]:
            if id(fp) not in active_param_ids and fp in _fused_optimizer.state:
                logging.debug("delete outdated params")
                del _fused_optimizer.state[fp]

        fused_group["params"] = new_params


def check_optimizer(optimizer):
    # make sure cloned attributes are not modified
    for attr in optimizer._bagua_cloned_attrs:
        if getattr(optimizer, attr) != getattr(optimizer._bagua_fused_optimizer, attr):
            logging.error(
                f"Should not change attribute {attr} in `optimizer.step(), maintain it in optimizer state.`"
            )


def sync_param_group_scalars(src_group, dst_group):
    for k, v in src_group.items():
        if k != "params":
            dst_group[k] = v

    for k, v in dst_group.items():
        if k not in src_group:
            dst_group[k] = None


def sync_optimizer_state(optimizer):
    # write back state for original params
    # Note: we should make sure every module parameter in original params groups has the right state
    _fused_optimizer = optimizer._bagua_fused_optimizer
    for group, fused_group in zip(
        optimizer.param_groups, _fused_optimizer.param_groups
    ):

        params = group["params"]
        fused_params = fused_group["params"]

        for fp in fused_params:
            if not hasattr(fp, "_bagua_fused_param_ids"):
                for name, v in _fused_optimizer.state[fp].items():
                    optimizer.state[fp][name] = v

            else:
                original_params = [params[i] for i in fp._bagua_fused_param_ids]

                for name, v in _fused_optimizer.state[fp].items():
                    if isinstance(v, torch.Tensor):
                        state_tensors = infer_state_tensors(
                            _fused_optimizer, fp, original_params, name
                        )

                        if state_tensors is not None:
                            for p, state in zip(original_params, state_tensors):
                                optimizer.state[p][name] = state
                    else:
                        for p in original_params:
                            optimizer.state[p][name] = v

                for p in original_params:
                    if len(optimizer.state[p]) != len(_fused_optimizer.state[fp]):
                        logging.warning("Something went wrong with optimizer state.")


def get_tensor_state(optimizer, param, name):
    s = optimizer.state[param][name]

    if not isinstance(s, torch.Tensor):
        logging.error(f"state {name} expected to be a tensor")

    return s


def get_optimizer_param_states(optimizer, params):
    state_tensors = {}  # type: Dict[str, List[torch.Tensor]]
    state_scalars = {}  # type: Dict[str, Any]

    state_tensor_names = set(
        [
            k
            for p in params
            for k, v in optimizer.state[p].items()
            if isinstance(v, torch.Tensor)
        ]
    )
    state_scalar_names = set(
        [
            k
            for p in params
            for k, v in optimizer.state[p].items()
            if not isinstance(v, torch.Tensor)
        ]
    )

    for name in state_tensor_names:
        tensors = []
        for p in params:
            if name not in optimizer.state[p]:
                logging.error(
                    f"Unexpected parameter state {name}, failed not fuse optimizer."
                )
                return None, None

            tensors.append(optimizer.state[p][name])

        state_tensors[name] = tensors

    for name in state_scalar_names:
        scalar = None

        for p in params:
            if name not in optimizer.state[p]:
                logging.error(
                    f"Unexpected parameter state {name}, failed not fuse optimizer."
                )
                return None, None

            if scalar is not None and scalar != optimizer.state[p][name]:
                logging.error(
                    f"Parameter state '{name}' does not match, failed not fuse optimizer."
                )
                return None, None

            state_scalars[name] = optimizer.state[p][name]

    return state_tensors, state_scalars



# Copyright (c) 2021 Kuaishou AI Platform & DS3 Lab
#
# All rights reserved.
#
# The file has been adapted from DeepSpeed:
#   https://github.com/microsoft/DeepSpeed/blob/master/deepspeed/moe/experts.py
# Git commit hash: bff6126f0ddbd1a03da66867571ac87b11c21ac1
# We retain the following license from the original files:

# Copyright 2020 The Microsoft DeepSpeed Team

import torch
import copy


class Experts(torch.nn.Module):
    def __init__(self, expert, num_local_experts=1):
        super(Experts, self).__init__()

        self.bagua_experts = torch.nn.ModuleList(
            [copy.deepcopy(expert) for i in range(num_local_experts)]
        )
        self.num_local_experts = num_local_experts

        # TODO: revisit allreduce for moe.gate...
        for expert in self.bagua_experts:
            # TODO: Create param groups to handle expert + data case (e.g. param.group = moe_group)
            for name, param in expert.named_parameters():
                param.expert = True

    def forward(self, inputs):
        chunks = inputs.chunk(self.num_local_experts, dim=1)
        expert_outputs = []
        for chunk, expert in zip(chunks, self.bagua_experts):
            out = expert(chunk)
            if type(out) is tuple:
                out = out[0]  # Ignore the bias term for now
            expert_outputs += [out]

        expert_output = torch.cat(expert_outputs, dim=1)
        return expert_output



import enum
from typing import List
import sys

if sys.version_info >= (3, 9):
    from typing import TypedDict  # pytype: disable=not-supported-yet
else:
    from typing_extensions import TypedDict  # pytype: disable=not-supported-yet
from pydantic import BaseModel


class TensorDtype(str, enum.Enum):
    F32 = "f32"
    F16 = "f16"
    U8 = "u8"


class TensorDeclaration(TypedDict):
    name: str
    num_elements: int
    dtype: TensorDtype


def get_tensor_declaration_bytes(td: TensorDeclaration) -> int:
    dtype_unit_size = {
        TensorDtype.F32.value: 4,
        TensorDtype.F16.value: 2,
        TensorDtype.U8.value: 1,
    }

    return td["num_elements"] * dtype_unit_size[td["dtype"]]


class BaguaHyperparameter(BaseModel):
    """
    Structured all bagua hyperparameters
    """

    buckets: List[List[TensorDeclaration]] = []
    bucket_size: int = 0
    is_hierarchical_reduce: bool = False

    def update(self, param_dict: dict):
        tmp = self.dict()
        tmp.update(param_dict)
        for key, value in param_dict.items():
            if key in tmp:
                self.__dict__[key] = value

        return self


class BaguaCoreTelemetrySpan(TypedDict):
    trace_id: int
    action: str
    tensor_name: str
    start_time: int
    end_time: int



# From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
# and IDIAP Research Institute nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# This file is copied from https://github.com/pytorch/pytorch/tree/v1.9.0/torch/testing/_core.py
"""
The testing package contains testing-specific utilities.
"""

import torch
import random
import math
import cmath
from typing import cast, List, Optional, Tuple, Union
import operator

FileCheck = torch._C.FileCheck

__all__ = [
    "FileCheck",
    "all_types",
    "all_types_and",
    "all_types_and_complex",
    "all_types_and_complex_and",
    "assert_allclose",
    "complex_types",
    "floating_and_complex_types",
    "floating_and_complex_types_and",
    "floating_types",
    "floating_types_and",
    "get_all_complex_dtypes",
    "get_all_dtypes",
    "get_all_device_types",
    "get_all_fp_dtypes",
    "get_all_int_dtypes",
    "get_all_math_dtypes",
    "integral_types",
    "integral_types_and",
    "make_non_contiguous",
    "rand_like",
    "randn_like",
]

rand_like = torch.rand_like
randn_like = torch.randn_like


# Helper function that returns True when the dtype is an integral dtype,
# False otherwise.
# TODO: implement numpy-like issubdtype
def is_integral(dtype: torch.dtype) -> bool:
    # Skip complex/quantized types
    dtypes = [x for x in get_all_dtypes() if x not in get_all_complex_dtypes()]
    return dtype in dtypes and not dtype.is_floating_point


def is_quantized(dtype: torch.dtype) -> bool:
    return dtype in (torch.quint8, torch.qint8, torch.qint32, torch.quint4x2)


# Helper function that maps a flattened index back into the given shape
# TODO: consider adding torch.unravel_index
def _unravel_index(flat_index, shape):
    flat_index = operator.index(flat_index)
    res = []

    # Short-circuits on zero dim tensors
    if shape == torch.Size([]):
        return 0

    for size in shape[::-1]:
        res.append(flat_index % size)
        flat_index = flat_index // size

    if len(res) == 1:
        return res[0]

    return tuple(res[::-1])


# (bool, msg) tuple, where msg is None if and only if bool is True.
_compare_return_type = Tuple[bool, Optional[str]]


# Compares two tensors with the same size on the same device and with the same
# dtype for equality.
# Returns a tuple (bool, msg). The bool value returned is True when the tensors
# are "equal" and False otherwise.
# The msg value is a debug string, and is None if the tensors are "equal."
# NOTE: Test Framework Tensor 'Equality'
#   Two tensors are "equal" if they are "close", in the sense of torch.allclose.
#   The only exceptions are complex tensors and bool tensors.
#
#   Complex tensors are "equal" if both the
#   real and complex parts (separately) are close. This is divergent from
#   torch.allclose's behavior, which compares the absolute values of the
#   complex numbers instead.
#
#   Using torch.allclose would be a less strict
#   comparison that would allow large complex values with
#   significant real or imaginary differences to be considered "equal,"
#   and would make setting rtol and atol for complex tensors distinct from
#   other tensor types.
#
#   Bool tensors are equal only if they are identical, regardless of
#   the rtol and atol values.
#
#   The `equal_nan` can be True or False, which maps to the True or False
#   in `torch.allclose`. `equal_nan` can also be "relaxed", which means
#   the complex will be compared in the relaxed mode:
#       2 + nan j == 3 + nan j ---> False when equal_nan=True
#                                   True when equal_nan="relaxed"
def _compare_tensors_internal(
    a: torch.Tensor, b: torch.Tensor, *, rtol, atol, equal_nan: Union[str, bool]
) -> _compare_return_type:
    assert equal_nan in {True, False, "relaxed"}
    debug_msg: Optional[str]
    # Integer (including bool) comparisons are identity comparisons
    # when rtol is zero and atol is less than one
    if (
        (is_integral(a.dtype) and rtol == 0 and atol < 1)
        or a.dtype is torch.bool
        or is_quantized(a.dtype)
    ):
        if (a == b).all().item():
            return (True, None)

        # Gathers debug info for failed integer comparison
        # NOTE: converts to long to correctly represent differences
        # (especially between uint8 tensors)
        identity_mask = a != b
        a_flat = a.to(torch.long).flatten()
        b_flat = b.to(torch.long).flatten()
        count_non_identical = torch.sum(identity_mask, dtype=torch.long)
        diff = torch.abs(a_flat - b_flat)
        greatest_diff_index = torch.argmax(diff)
        debug_msg = (
            "Found {0} different element(s) (out of {1}), with the greatest "
            "difference of {2} ({3} vs. {4}) occuring at index "
            "{5}.".format(
                count_non_identical.item(),
                a.numel(),
                diff[greatest_diff_index],
                a_flat[greatest_diff_index],
                b_flat[greatest_diff_index],
                _unravel_index(greatest_diff_index, a.shape),
            )
        )
        return (False, debug_msg)

    # Compares complex tensors' real and imaginary parts separately.
    # (see NOTE Test Framework Tensor "Equality")
    if a.is_complex():
        if equal_nan == "relaxed":
            a = a.clone()
            b = b.clone()
            a.real[a.imag.isnan()] = math.nan
            a.imag[a.real.isnan()] = math.nan
            b.real[b.imag.isnan()] = math.nan
            b.imag[b.real.isnan()] = math.nan

        real_result, debug_msg = _compare_tensors_internal(
            a.real, b.real, rtol=rtol, atol=atol, equal_nan=equal_nan
        )

        if not real_result:
            debug_msg = "Real parts failed to compare as equal! " + cast(str, debug_msg)
            return (real_result, debug_msg)

        imag_result, debug_msg = _compare_tensors_internal(
            a.imag, b.imag, rtol=rtol, atol=atol, equal_nan=equal_nan
        )

        if not imag_result:
            debug_msg = "Imaginary parts failed to compare as equal! " + cast(
                str, debug_msg
            )
            return (imag_result, debug_msg)

        return (True, None)

    # All other comparisons use torch.allclose directly
    if torch.allclose(
        a, b, rtol=rtol, atol=atol, equal_nan=(equal_nan in {"relaxed", True})
    ):
        return (True, None)

    # Gathers debug info for failed float tensor comparison
    # NOTE: converts to float64 to best represent differences
    a_flat = a.to(torch.float64).flatten()
    b_flat = b.to(torch.float64).flatten()
    diff = torch.abs(a_flat - b_flat)

    # Masks close values
    # NOTE: this avoids (inf - inf) oddities when computing the difference
    close = torch.isclose(a_flat, b_flat, rtol, atol, (equal_nan in {"relaxed", True}))
    diff[close] = 0
    nans = torch.isnan(diff)
    num_nans = nans.sum()

    outside_range = (diff > (atol + rtol * torch.abs(b_flat))) | (diff == math.inf)
    count_outside_range = torch.sum(outside_range, dtype=torch.long)
    greatest_diff_index = torch.argmax(diff)
    debug_msg = (
        "With rtol={0} and atol={1}, found {2} element(s) (out of {3}) whose "
        "difference(s) exceeded the margin of error (including {4} nan comparisons). "
        "The greatest difference was {5} ({6} vs. {7}), which "
        "occurred at index {8}.".format(
            rtol,
            atol,
            count_outside_range + num_nans,
            a.numel(),
            num_nans,
            diff[greatest_diff_index],
            a_flat[greatest_diff_index],
            b_flat[greatest_diff_index],
            _unravel_index(greatest_diff_index, a.shape),
        )
    )
    return (False, debug_msg)


# Checks if two scalars are equal(-ish), returning (True, None)
# when they are and (False, debug_msg) when they are not.
def _compare_scalars_internal(
    a, b, *, rtol: float, atol: float, equal_nan: Union[str, bool]
) -> _compare_return_type:
    def _helper(a, b, s) -> _compare_return_type:
        # Short-circuits on identity
        if a == b or ((equal_nan in {"relaxed", True}) and a != a and b != b):
            return (True, None)

        # Special-case for NaN comparisions when equal_nan=False
        if not (equal_nan in {"relaxed", True}) and (a != a or b != b):
            msg = (
                "Found {0} and {1} while comparing" + s + "and either one "
                "is nan and the other isn't, or both are nan and "
                "equal_nan is False"
            ).format(a, b)
            return (False, msg)

        diff = abs(a - b)
        allowed_diff = atol + rtol * abs(b)
        result = diff <= allowed_diff

        # Special-case for infinity comparisons
        # NOTE: if b is inf then allowed_diff will be inf when rtol is not 0
        if (math.isinf(a) or math.isinf(b)) and a != b:
            result = False

        msg = None
        if not result:
            msg = (
                "Comparing" + s + "{0} and {1} gives a "
                "difference of {2}, but the allowed difference "
                "with rtol={3} and atol={4} is "
                "only {5}!"
            ).format(a, b, diff, rtol, atol, allowed_diff)

        return result, msg

    if isinstance(a, complex) or isinstance(b, complex):
        a = complex(a)
        b = complex(b)

        if equal_nan == "relaxed":
            if cmath.isnan(a) and cmath.isnan(b):
                return (True, None)

        result, msg = _helper(a.real, b.real, " the real part ")

        if not result:
            return (False, msg)

        return _helper(a.imag, b.imag, " the imaginary part ")

    return _helper(a, b, " ")


def assert_allclose(
    actual, expected, rtol=None, atol=None, equal_nan=True, msg=""
) -> None:
    if not isinstance(actual, torch.Tensor):
        actual = torch.tensor(actual)
    if not isinstance(expected, torch.Tensor):
        expected = torch.tensor(expected, dtype=actual.dtype)
    if expected.shape != actual.shape:
        raise AssertionError(
            "expected tensor shape {0} doesn't match with actual tensor "
            "shape {1}!".format(expected.shape, actual.shape)
        )
    if rtol is None or atol is None:
        if rtol is not None or atol is not None:
            raise ValueError(
                "rtol and atol must both be specified or both be unspecified"
            )
        rtol, atol = _get_default_tolerance(actual, expected)

    result, debug_msg = _compare_tensors_internal(
        actual, expected, rtol=rtol, atol=atol, equal_nan=equal_nan
    )

    if result:
        return

    if msg is None or msg == "":
        msg = debug_msg

    raise AssertionError(msg)


def make_non_contiguous(tensor: torch.Tensor) -> torch.Tensor:
    if tensor.numel() <= 1:  # can't make non-contiguous
        return tensor.clone()
    osize = list(tensor.size())

    # randomly inflate a few dimensions in osize
    for _ in range(2):
        dim = random.randint(0, len(osize) - 1)
        add = random.randint(4, 15)
        osize[dim] = osize[dim] + add

    # narrow doesn't make a non-contiguous tensor if we only narrow the 0-th dimension,
    # (which will always happen with a 1-dimensional tensor), so let's make a new
    # right-most dimension and cut it off

    input = tensor.new(torch.Size(osize + [random.randint(2, 3)]))
    input = input.select(len(input.size()) - 1, random.randint(0, 1))
    # now extract the input of correct size from 'input'
    for i in range(len(osize)):
        if input.size(i) != tensor.size(i):
            bounds = random.randint(1, input.size(i) - tensor.size(i))
            input = input.narrow(i, bounds, tensor.size(i))

    input.copy_(tensor)

    # Use .data here to hide the view relation between input and other temporary Tensors
    return input.data


# Functions and classes for describing the dtypes a function supports
# NOTE: these helpers should correspond to PyTorch's C++ dispatch macros

# Verifies each given dtype is a torch.dtype
def _validate_dtypes(*dtypes):
    for dtype in dtypes:
        assert isinstance(dtype, torch.dtype)
    return dtypes


# class for tuples corresponding to a PyTorch dispatch macro
class _dispatch_dtypes(tuple):
    def __add__(self, other):
        assert isinstance(other, tuple)
        return _dispatch_dtypes(tuple.__add__(self, other))


_floating_types = _dispatch_dtypes((torch.float32, torch.float64))


def floating_types():
    return _floating_types


_floating_types_and_half = _floating_types + (torch.half,)


def floating_types_and_half():
    return _floating_types_and_half


def floating_types_and(*dtypes):
    return _floating_types + _validate_dtypes(*dtypes)


_floating_and_complex_types = _floating_types + (torch.cfloat, torch.cdouble)


def floating_and_complex_types():
    return _floating_and_complex_types


def floating_and_complex_types_and(*dtypes):
    return _floating_and_complex_types + _validate_dtypes(*dtypes)


_integral_types = _dispatch_dtypes(
    (torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64)
)


def integral_types():
    return _integral_types


def integral_types_and(*dtypes):
    return _integral_types + _validate_dtypes(*dtypes)


_all_types = _floating_types + _integral_types


def all_types():
    return _all_types


def all_types_and(*dtypes):
    return _all_types + _validate_dtypes(*dtypes)


_complex_types = _dispatch_dtypes((torch.cfloat, torch.cdouble))


def complex_types():
    return _complex_types


_all_types_and_complex = _all_types + _complex_types


def all_types_and_complex():
    return _all_types_and_complex


def all_types_and_complex_and(*dtypes):
    return _all_types_and_complex + _validate_dtypes(*dtypes)


_all_types_and_half = _all_types + (torch.half,)


def all_types_and_half():
    return _all_types_and_half


def get_all_dtypes(
    include_half=True,
    include_bfloat16=True,
    include_bool=True,
    include_complex=True,
    include_complex32=False,
) -> List[torch.dtype]:
    dtypes = get_all_int_dtypes() + get_all_fp_dtypes(
        include_half=include_half, include_bfloat16=include_bfloat16
    )
    if include_bool:
        dtypes.append(torch.bool)
    if include_complex:
        dtypes += get_all_complex_dtypes(include_complex32)
    return dtypes


def get_all_math_dtypes(device) -> List[torch.dtype]:
    return (
        get_all_int_dtypes()
        + get_all_fp_dtypes(
            include_half=device.startswith("cuda"), include_bfloat16=False
        )
        + get_all_complex_dtypes()
    )


def get_all_complex_dtypes(include_complex32=False) -> List[torch.dtype]:
    return (
        [torch.complex32, torch.complex64, torch.complex128]
        if include_complex32
        else [torch.complex64, torch.complex128]
    )


def get_all_int_dtypes() -> List[torch.dtype]:
    return [torch.uint8, torch.int8, torch.int16, torch.int32, torch.int64]


def get_all_fp_dtypes(include_half=True, include_bfloat16=True) -> List[torch.dtype]:
    dtypes = [torch.float32, torch.float64]
    if include_half:
        dtypes.append(torch.float16)
    if include_bfloat16:
        dtypes.append(torch.bfloat16)
    return dtypes


def get_all_device_types() -> List[str]:
    return ["cpu"] if not torch.cuda.is_available() else ["cpu", "cuda"]


# 'dtype': (rtol, atol)
_default_tolerances = {
    "float64": (1e-5, 1e-8),  # NumPy default
    "float32": (1e-4, 1e-5),  # This may need to be changed
    "float16": (1e-3, 1e-3),  # This may need to be changed
}


def _get_default_tolerance(a, b=None) -> Tuple[float, float]:
    if b is None:
        dtype = str(a.dtype).split(".")[-1]  # e.g. "float32"
        return _default_tolerances.get(dtype, (0, 0))
    a_tol = _get_default_tolerance(a)
    b_tol = _get_default_tolerance(b)
    return (max(a_tol[0], b_tol[0]), max(a_tol[1], b_tol[1]))



import torch


class MinMaxUInt8:
    def __init__(self):
        self.eps = 1e-7
        self.quantization_level = 255.0

    def compress(self, tensor: torch.Tensor) -> (torch.Tensor, torch.Tensor):
        _min = torch.min(tensor)
        _max = torch.max(tensor)
        scale = self.quantization_level / (_max - _min + self.eps)
        upper_bound = torch.round(_max * scale)
        lower_bound = upper_bound - self.quantization_level

        level = torch.round(tensor * scale)
        level = torch.clamp(level, max=upper_bound)

        _minmax = torch.zeros(2, dtype=tensor.dtype, device=tensor.device)
        _minmax[0] = _min
        _minmax[1] = _max
        return _minmax, (level - lower_bound).to(torch.uint8)

    def decompress(
        self, _minmax: torch.Tensor, compressed: torch.Tensor
    ) -> torch.Tensor:
        _min = _minmax[0]
        _max = _minmax[1]

        scale = self.quantization_level / (_max - _min + self.eps)
        upper_bound = torch.round(_max * scale)
        lower_bound = upper_bound - self.quantization_level
        return (compressed.float() + lower_bound) / scale


if __name__ == "__main__":
    x = torch.rand(100).cuda()
    _minmax, compressed = MinMaxUInt8().compress(x)
    decompressed = MinMaxUInt8().decompress(_minmax, compressed)

    diff = x - decompressed
    print(f"{diff}, {torch.norm(diff)}")



import collections
import tempfile
import copy
import math
import logging
import csv
from typing import Tuple, List, Dict

from .bayesian_optimizer import (
    IntParam,
    BoolParam,
    BayesianOptimizer,
)
from bagua.bagua_define import (
    TensorDtype,
    TensorDeclaration,
    BaguaHyperparameter,
)


class AutotuneTaskManager:
    RECORD_MAX_NUM = 1000

    def __init__(
        self,
        task_name,
        need_to_log: bool,
    ) -> None:
        self.task_name = task_name
        self.record_deque = collections.deque(
            [
                (
                    -1,
                    BaguaHyperparameter(),
                    float("-inf"),
                )
            ]
        )
        if need_to_log:
            tmpfile = tempfile.NamedTemporaryFile(
                prefix="bagua_autotune_", mode="w", suffix=".log", delete=False
            )
            self.autotune_logfile_path = tmpfile.name
            tmpfile.close()
        else:
            self.autotune_logfile_path = None

        self.bayesian_optimizer = BayesianOptimizer(
            {
                "bucket_size_2p": IntParam(  # bucket_size = 2 ^ bucket_size_2p
                    val=13,
                    space_dimension=(  # 1KB ~ 2GB
                        10,
                        31,
                    ),
                ),
                "is_hierarchical_reduce": BoolParam(False),
            },
        )

    @staticmethod
    def record_autotune_log(
        autotune_logfile_path: str, autotune_hp: dict, train_iter: int, score: float
    ):
        with open(autotune_logfile_path, "a") as autotune_log:
            csv_writer = csv.DictWriter(
                autotune_log,
                fieldnames=sorted(["train_iter", "score"] + list(autotune_hp.keys())),
            )
            first_line = open(autotune_logfile_path).readline()
            if not first_line:
                csv_writer.writeheader()

            cols = copy.deepcopy(autotune_hp)
            cols.update(
                {
                    "train_iter": train_iter,
                    "score": score,
                }
            )
            cols = collections.OrderedDict(cols)
            logging.info("cols={}".format(cols))
            csv_writer.writerow(cols)

    @staticmethod
    def split_bucket_by_bucket_size(
        tensor_list: List[TensorDeclaration],
        bucket_size: int,
        param_group_info: Dict[str, int] = {},
    ):
        dtype_unit_size = {
            TensorDtype.F32.value: 4,
            TensorDtype.F16.value: 2,
            TensorDtype.U8.value: 1,
        }

        buckets = []
        tmp_bucket = []
        tmp_bucket_size = 0
        for (dtype, unit_size) in sorted(dtype_unit_size.items()):
            for tensor_declar in [x for x in tensor_list if x["dtype"] == dtype]:
                tensor_size = tensor_declar["num_elements"] * unit_size
                tmp_bucket_size += tensor_size
                tmp_bucket.append(tensor_declar)
                if tmp_bucket_size >= bucket_size:
                    buckets.append(tmp_bucket)
                    tmp_bucket = []
                    tmp_bucket_size = 0

        if len(tmp_bucket) != 0:
            buckets.append(tmp_bucket)

        # group params in bucket
        for i, _ in enumerate(buckets):
            buckets[i] = sorted(
                buckets[i], key=lambda p: param_group_info.get(p["name"], -1)
            )

        return buckets

    def tail_record(self) -> Tuple[int, BaguaHyperparameter, float]:
        return self.record_deque[-1]

    def best_hyperparameter(self) -> BaguaHyperparameter:
        return sorted(
            [(score, hp) for (_, hp, score) in self.record_deque],
            key=lambda pair: pair[0],
        )[-1][1]

    def report_metrics(
        self,
        train_iter: int,
        hyperparameter: BaguaHyperparameter,
        system_efficiency_score: float,
    ) -> None:
        while len(self.record_deque) > self.RECORD_MAX_NUM:
            self.record_deque.pop()
        self.record_deque.append(
            (
                train_iter,
                hyperparameter,
                system_efficiency_score,
            )
        )

    def ask_hyperparmeter(
        self,
        train_iter: int,
        tensor_partial_order: Dict[str, int] = {},  # tensor_name -> rank
    ) -> BaguaHyperparameter:
        (_, hp, system_efficiency_score) = self.tail_record()
        optimizer_params = {
            "bucket_size_2p": int(math.log(hp.bucket_size, 2)),
            "is_hierarchical_reduce": hp.is_hierarchical_reduce,
        }
        self.bayesian_optimizer.tell(optimizer_params, system_efficiency_score)
        recommend_param = self.bayesian_optimizer.ask()
        recommend_bucket_size = 2 ** recommend_param["bucket_size_2p"]

        if self.autotune_logfile_path:
            AutotuneTaskManager.record_autotune_log(
                self.autotune_logfile_path,
                optimizer_params,
                train_iter,
                system_efficiency_score,
            )
        tensor_list = [
            tensor_declar for bucket in hp.buckets for tensor_declar in bucket
        ]
        tensor_list = sorted(
            tensor_list, key=lambda td: tensor_partial_order.get(td["name"], -1)
        )

        recommend_buckets = AutotuneTaskManager.split_bucket_by_bucket_size(
            tensor_list,
            recommend_bucket_size,
        )

        recommend_hp = BaguaHyperparameter(
            buckets=recommend_buckets,
            bucket_size=recommend_bucket_size,
            is_hierarchical_reduce=bool(recommend_param["is_hierarchical_reduce"]),
        )

        return recommend_hp



#!/usr/bin/env python3

from .base import Algorithm, AlgorithmImpl, GlobalAlgorithmRegistry  # noqa: F401
from . import bytegrad, decentralized, gradient_allreduce  # noqa: F401
from . import q_adam, async_model_average  # noqa: F401


GlobalAlgorithmRegistry.register(
    "gradient_allreduce",
    gradient_allreduce.GradientAllReduceAlgorithm,
    description="Gradient AllReduce Algorithm",
)
GlobalAlgorithmRegistry.register(
    "bytegrad", bytegrad.ByteGradAlgorithm, description="ByteGrad Algorithm"
)
GlobalAlgorithmRegistry.register(
    "decentralized",
    decentralized.DecentralizedAlgorithm,
    description="Decentralized SGD Algorithm",
)
GlobalAlgorithmRegistry.register(
    "low_precision_decentralized",
    decentralized.LowPrecisionDecentralizedAlgorithm,
    description="Low Precision Decentralized SGD Algorithm",
)
GlobalAlgorithmRegistry.register(
    "qadam", q_adam.QAdamAlgorithm, description="QAdam Algorithm"
)
GlobalAlgorithmRegistry.register(
    "async",
    async_model_average.AsyncModelAverageAlgorithm,
    description="Asynchronous Model Average Algorithm",
)



# Copyright (c) 2021 Kuaishou AI Platform & DS3 Lab
#
# All rights reserved.
#
# The file has been adapted from DeepSpeed:
#   https://github.com/microsoft/DeepSpeed/blob/master/deepspeed/moe/layer.py
# Git commit hash: bff6126f0ddbd1a03da66867571ac87b11c21ac1
# We retain the following license from the original files:

# Copyright 2020 The Microsoft DeepSpeed Team

import bagua.torch_api as bagua
import logging
import torch
import torch.distributed as dist

from .sharded_moe import MOELayer, TopKGate
from .experts import Experts
import typing


class MoE(torch.nn.Module):
    def __init__(
        self,
        hidden_size,
        expert,
        num_local_experts=1,
        k=1,
        output_dropout_prob=0.0,
        capacity_factor=1.0,
        eval_capacity_factor=1.0,
        min_capacity=4,
        noisy_gate_policy: typing.Optional[str] = None,
    ):
        """Initialize an MoE layer.

        Arguments:
            hidden_size (int): the hidden dimension of the model, importantly this is also the input and output dimension.

            expert (torch.nn.Module): the torch module that defines the expert (e.g., MLP, torch.linear).

            num_local_experts (int, optional): default=1, number of local experts per gpu.

            k (int, optional): default=1, top-k gating value, only supports k=1 or k=2.

            output_dropout_prob (float, optional): default=0.0, output dropout probability.

            capacity_factor (float, optional): default=1.0, the capacity of the expert at training time.

            eval_capacity_factor (float, optional): default=1.0, the capacity of the expert at eval time.

            min_capacity (int, optional): default=4, the minimum capacity per expert regardless of the capacity_factor.

            noisy_gate_policy (str, optional): default=None, noisy gate policy, valid options are 'Jitter', 'RSample' or 'None'.
        """

        super(MoE, self).__init__()

        assert noisy_gate_policy is None or noisy_gate_policy in [
            "None",
            "Jitter",
            "RSample",
        ], (
            "Unsupported noisy_gate_policy: " + noisy_gate_policy
        )

        self.num_experts = num_local_experts * bagua.get_world_size()
        logging.info(
            f"num_experts: {self.num_experts} | num_local_experts: {num_local_experts} | world_size: {bagua.get_world_size()}"
        )

        experts = Experts(expert, num_local_experts)
        self.bagua_moe = MOELayer(
            TopKGate(
                hidden_size,
                self.num_experts,
                k,
                capacity_factor,
                eval_capacity_factor,
                min_capacity,
                noisy_gate_policy,
            ),
            experts,
            num_local_experts,
            group=dist.group.WORLD,
        )

        self.dropout = torch.nn.Dropout(output_dropout_prob)

    def forward(self, hidden_states, used_token=None):
        """MoE forward

        Arguments:
            hidden_states (Tensor): input to the layer
            used_token (Tensor, optional): default: None, mask only used tokens

        Returns:
            A tuple including output, gate loss, and expert count.

            * output (Tensor): output of the model

            * l_aux (Tensor): gate loss value

            * exp_counts (int): expert count
        """
        output = self.bagua_moe(hidden_states, used_token)
        output = self.dropout(output)
        return output, self.bagua_moe.l_aux, self.bagua_moe.exp_counts



import unittest
from bagua.bagua_define import (
    TensorDeclaration,
    BaguaHyperparameter,
    TensorDtype,
)
from tests import skip_if_cuda_available


class TestBaguaDefine(unittest.TestCase):
    @skip_if_cuda_available()
    def test_bagua_hyperparameter(self):
        hp = BaguaHyperparameter(is_hierarchical_reduce=False)

        buckets = [
            [
                TensorDeclaration(
                    {"name": "A", "num_elements": 123, "dtype": TensorDtype.F32}
                )
            ]
        ]

        hp = hp.update(
            {
                "buckets": buckets,
            }
        )

        assert hp.buckets == buckets
        assert hp.is_hierarchical_reduce is False


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
from bagua.torch_api.communication import get_backend
from typing import Optional, Callable

import torch
import bagua_core as B
import gorilla


@gorilla.patches(torch.Tensor, filter=lambda name, obj: "bagua" in name)
class BaguaTensor:
    """
    This class patch `torch.Tensor <https://pytorch.org/docs/stable/tensors.html?highlight=tensor#torch.Tensor>`_
    with additional methods.

    A Bagua tensor is required to use Bagua's communication algorithms. Users can convert a PyTorch tensor to Bagua
    tensor by :meth:`ensure_bagua_tensor`.

    Bagua tensor features a proxy structure, where the actual tensor used by backend is accessed via a **"Proxy Tensor"**.
    The proxy tensor is registered in Bagua, whenever the Bagua backend needs a tensor (for example use it for
    communication), it calls the :meth:`bagua_getter_closure` on the proxy tensor to get the tensor that is actually
    worked on. We call this tensor **"Effective Tensor"**. The :attr:`bagua_setter_closure` is also provided to replace
    the effective tensor during runtime. It is intended to be used to replace the effective tensor with customized
    workflow.

    Their relation can be seen in the following diagram:

    .. image:: https://user-images.githubusercontent.com/18649508/139179394-51d0c0f5-e233-4ada-8e5e-0e70a889540d.png

    For example, in the gradient allreduce algorithm, the effective tensor that needs to be exchanged between
    machines is the gradient.  In this case, we will register the model parameters as proxy tensor, and register
    :meth:`bagua_getter_closure` to be ``lambda proxy_tensor: proxy_tensor.grad``. In this way, even if the gradient
    tensor is recreated or changed during runtime, Bagua can still identify the correct tensor and use it
    for communication, since the proxy tensor serves as the root for access and is never replaced.
    """

    def _bagua_sanity_check(self):
        assert (
            self._bagua_backend_tensor.data_ptr()
            == self.bagua_getter_closure().data_ptr()
        )
        assert (
            self._bagua_backend_tensor.num_elements()
            == self.bagua_getter_closure().numel()
        )
        assert (
            self._bagua_backend_tensor.num_elements_allocated()
            == self.bagua_getter_closure().numel()
        )

    def is_bagua_tensor(self) -> bool:
        """
        Checking if this is a Bagua tensor.
        """
        return hasattr(self, "_bagua_backend_tensor")

    def ensure_bagua_tensor(
        self,
        name: Optional[str] = None,
        module_name: Optional[str] = None,
        getter_closure: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        setter_closure: Optional[Callable[[torch.Tensor, torch.Tensor], None]] = None,
    ):
        """
        Convert a PyTorch tensor or parameter to Bagua tensor inplace and return it.
        A Bagua tensor is required to use Bagua's communication algorithms.

        This operation will register :attr:`self` as proxy tensor to the Bagua backend. :attr:`getter_closure` takes
        the proxy tensor as input and returns a Pytorch tensor. When using the Bagua tensor, the :attr:`getter_closure`
        will be called and returns the effective tensor which will be used for communication and other operations.
        For example, if one of a model's parameter ``param`` is registered as proxy tensor, and :attr:`getter_closure`
        is ``lambda x: x.grad``, during runtime its gradient will be used.

        :attr:`setter_closure` takes the proxy tensor and another tensor as inputs and returns nothing.
        It is mainly used for changing the effective tensor used in runtime. For example when one of
        a model's parameter ``param`` is registered as proxy tensor, and :attr:`getter_closure` is ``lambda x: x.grad``,
        the :attr:`setter_closure` can be ``lambda param, new_grad_tensor: setattr(param, "grad", new_grad_tensor)``.
        When the :attr:`setter_closure` is called, the effective tensor used in later operations will be changed
        to ``new_grad_tensor``.

        Args:
            name: The unique name of the tensor.
            module_name: The name of the model of which the tensor belongs to.
              The model name can be acquired using ``model.bagua_module_name``.
              This is required to call :meth:`bagua_mark_communication_ready` related methods.
            getter_closure: A function that accepts a Pytorch tensor as its input and returns a Pytorch tensor as
              its output. Could be ``None``, which means an identity mapping ``lambda x: x`` is used. Default: ``None``.
            setter_closure: A function that accepts two Pytorch tensors as its inputs and returns nothing. Could be ``None``,
              which is a no-op. Default: ``None``.

        Returns:
            The original tensor with Bagua tensor attributes initialized.
        """
        if self.is_bagua_tensor():
            if name is not None:
                assert (
                    self.bagua_tensor_name == name
                ), "assigning a different name to existing bagua tensor is forbidden"
                ", self.bagua_tensor_name={}, name={}".format(self.bagua_tensor_name, name)

                assert (
                    self.bagua_module_name == module_name
                ), "assigning a different module name to existing bagua tensor is forbidden"
                ", self.bagua_module_name={}, module_name={}".format(self.bagua_module_name, module_name)

            if (
                getter_closure == self._bagua_getter_closure
                and setter_closure == self._bagua_setter_closure
            ):
                return self

        self.bagua_tensor_name = name if name is not None else ""
        self.bagua_module_name = module_name
        self.bagua_backend = (
            get_backend(self.bagua_module_name)
            if self.bagua_module_name is not None
            else None
        )

        # initialize backend tensor
        if setter_closure is not None:
            assert (
                getter_closure is not None
            ), "must provide `getter_closure` when `setter_closure` is not None"
        self._bagua_getter_closure = getter_closure
        self._bagua_setter_closure = setter_closure

        self._bagua_backend_tensor = B.BaguaTensorPy(
            name=self.bagua_tensor_name,
            torch_tensor=self,
        )

        self._bagua_sanity_check()

        self._bagua_ready_event = torch.cuda.Event()
        self._bagua_bucket = None
        return self

    def to_bagua_tensor(
        self,
        name: Optional[str] = None,
        module_name: Optional[str] = None,
        getter_closure: Optional[Callable[[torch.Tensor], torch.Tensor]] = None,
        setter_closure: Optional[Callable[[torch.Tensor, torch.Tensor], None]] = None,
    ):
        """
        Create a new Bagua tensor from a PyTorch tensor or parameter and return it.
        The new Bagua tensor will share the same storage with the input PyTorch tensor.
        A Bagua tensor is required to use Bagua's communication algorithms.
        See :meth:`ensure_bagua_tensor` for more information.

        Caveat: Be aware that if the original tensor changes to use a different storage
        using for example ``torch.Tensor.set_(...)``, the new Bagua tensor will still
        use the old storage.

        Args:
            name: The unique name of the tensor.
            module_name: The name of the model of which the tensor belongs to.
              The model name can be acquired using ``model.bagua_module_name``.
              This is required to call :meth:`bagua_mark_communication_ready` related methods.
            getter_closure: A function that accepts a Pytorch tensor as its input and returns a Pytorch tensor as
              its output. See :meth:`ensure_bagua_tensor`.
            setter_closure: A function that accepts two Pytorch tensors as its inputs and returns nothing. See :meth:`ensure_bagua_tensor`.
        Returns:
            The new Bagua tensor sharing the same storage with the original tensor.
        """
        new_tensor = self.view(self.dtype)
        return new_tensor.ensure_bagua_tensor(
            name, module_name, getter_closure, setter_closure
        )

    def bagua_getter_closure(self) -> torch.Tensor:
        """Returns the tensor that will be used in runtime."""
        return (
            self._bagua_getter_closure(self)
            if self._bagua_getter_closure is not None
            else self
        )

    def bagua_setter_closure(self, tensor: torch.Tensor):
        """Sets the tensor that will be used in runtime to a new Pytorch tensor :attr:`tensor`.

        Args:
            tensor: The new tensor to be set to.
        """

        assert self._bagua_setter_closure is not None
        self._bagua_setter_closure(self, tensor)

    def bagua_backend_tensor(self) -> B.BaguaTensorPy:
        """
        Returns:
            The raw Bagua backend tensor.
        """
        return self._bagua_backend_tensor

    def bagua_ensure_grad(self) -> torch.Tensor:
        """
        Create a zero gradient tensor for the current parameter if not exist.

        Returns:
            The original tensor.
        """
        if hasattr(self, "grad") and self.grad is not None:
            return self
        elif isinstance(self, torch.nn.Parameter):
            with torch.no_grad():
                t = torch.zeros_like(self.data)
                self.grad = t
            return self
        else:
            raise NotImplementedError

    def bagua_mark_communication_ready(self):
        """
        Mark a Bagua tensor ready for scheduled operations execution.
        """
        torch.cuda.current_stream().record_event(self._bagua_ready_event)
        assert (
            self.bagua_backend is not None
        ), "tensor must be initialized with module name to call mark ready"
        self.bagua_backend.mark_communication_ready(
            self.bagua_backend_tensor(),
            self._bagua_ready_event.cuda_event,
        )

    def bagua_mark_communication_ready_without_synchronization(self):
        """
        Mark a Bagua tensor ready immediately, without `CUDA event <https://pytorch.org/docs/stable/generated/torch.cuda.Event.html?highlight=event#torch.cuda.Event>`_ synchronization.
        """
        assert (
            self.bagua_backend is not None
        ), "tensor must be initialized with module name to call mark ready"
        self.bagua_backend.mark_communication_ready(
            self.bagua_backend_tensor(),
            0,
        )

    def bagua_set_storage(
        self,
        storage: torch.Storage,
        storage_offset: int = 0,
    ):
        """
        Sets the underlying storage for the effective tensor returned by :meth:`bagua_getter_closure` with an existing
        `torch.Storage <https://pytorch.org/docs/stable/storage.html?highlight=storage>`_.

        Args:
            storage: The storage to use.
            storage_offset: The offset in the storage.
        """
        if self._bagua_setter_closure is None:
            # set directly
            with torch.no_grad():
                self.bagua_getter_closure().set_(
                    storage, storage_offset, self.bagua_getter_closure().shape
                )
            return

        with torch.no_grad():
            t = torch.zeros_like(self.bagua_getter_closure())
            t.set_(storage, storage_offset, t.shape)
            self.bagua_setter_closure(t)


_base = gorilla._get_base(BaguaTensor)
_decorator_data = gorilla.get_decorator_data(_base)
for patch in _decorator_data.patches:
    gorilla.apply(patch)



# Copyright (c) Uber Technologies, Inc. and its affiliates.
# Copyright (c) 2021 Kuaishou AI Platform & DS3 Lab.
#
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from distutils.version import LooseVersion

import torch
from torch.autograd.function import Function
import torch.nn.functional as F
from torch.nn.modules.batchnorm import _BatchNorm

import bagua.torch_api as bagua
from bagua.torch_api.communication import allgather, allreduce

# Backward compat for old PyTorch
if not hasattr(torch.jit, "unused"):
    torch.jit.unused = lambda x: x


_SYNC_BN_V2 = LooseVersion(torch.__version__) >= LooseVersion("1.5.0") and LooseVersion(
    torch.__version__
) <= LooseVersion("1.6.0")
_SYNC_BN_V3 = LooseVersion(torch.__version__) >= LooseVersion("1.6.0")
_SYNC_BN_V4 = LooseVersion(torch.__version__) >= LooseVersion("1.9.0")


class SyncBatchNorm(_BatchNorm):
    r"""Applies synchronous BatchNorm for distributed module with N-dimensional BatchNorm layer(s).
    See `BatchNorm <https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm2d.html?highlight=batchnorm#torch.nn.BatchNorm2d>`_ for more details.

    Arguments:
        num_features: Number of channels :math:`C` from the shape :math:`(N, C, ...)`.
        eps: A value added to the denominator for numerical stability. Default: 1e-5.
        momentum: The value used for the running_mean and running_var
            computation. Can be set to ``None`` for cumulative moving average
            (i.e. simple average). Default: 0.1.
        affine: A boolean value that when set to ``True``, this module has
            learnable affine parameters. Default: ``True``.
        track_running_stats: A boolean value that when set to ``True``, this
            module tracks the running mean and variance, and when set to ``False``,
            this module does not track such statistics and always uses batch
            statistics in both training and eval modes. Default: ``True``.

    .. note:: Only GPU input tensors are supported in the training mode.
    """

    def __init__(
        self,
        num_features,
        eps=1e-5,
        momentum=0.1,
        affine=True,
        track_running_stats=True,
    ):
        super().__init__(num_features, eps, momentum, affine, track_running_stats)

    def _check_input_dim(self, input):
        if input.dim() < 2:
            raise ValueError(
                "expected at least 2D input (got {}D input)".format(input.dim())
            )

    def _run_bn(self, input):
        return F.batch_norm(
            input,
            self.running_mean,
            self.running_var,
            self.weight,
            self.bias,
            self.training or not self.track_running_stats,
            self.momentum,
            self.eps,
        )

    @torch.jit.unused
    def _maybe_run_sync_bn(self, input):
        if bagua.get_world_size() == 1:
            return self._run_bn(input)
        return _SyncBatchNorm.apply(
            input,
            self.weight,
            self.bias,
            self.running_mean,
            self.running_var,
            self.eps,
            self.momentum,
        )

    def forward(self, input):
        # currently only GPU input is supported by underlying kernel from PyTorch
        if not input.is_cuda:
            raise ValueError("SyncBatchNorm expected input tensor to be on GPU")

        self._check_input_dim(input)

        if self.training and self.track_running_stats:
            assert self.num_batches_tracked is not None
            self.num_batches_tracked = self.num_batches_tracked + 1

        if not self.training and self.track_running_stats:
            return self._run_bn(input)
        else:
            return self._maybe_run_sync_bn(input)

    @classmethod
    def convert_sync_batchnorm(cls, module):
        r"""Helper function to convert all :attr:`BatchNorm*D` layers in the model to
        `torch.nn.SyncBatchNorm <https://pytorch.org/docs/stable/generated/torch.nn.SyncBatchNorm.html?highlight=syncbatchnorm#torch.nn.SyncBatchNorm>`_ layers.

        Arguments:
            module (nn.Module): Module containing one or more :attr:`BatchNorm*D` layers

        Returns:
            The original :attr:`module` with the converted :class:`torch.nn.SyncBatchNorm`
            layers. If the original :attr:`module` is a :attr:`BatchNorm*D` layer,
            a new :class:`torch.nn.SyncBatchNorm` layer object will be returned
            instead.

        .. note:: This function must be called before :meth:`~bagua.torch_api.distributed.BaguaModule.with_bagua` method.

        Example::
            >>> # Network with nn.BatchNorm layer
            >>> model = torch.nn.Sequential(
            ...      torch.nn.Linear(D_in, H),
            ...      torch.nn.ReLU(),
            ...      torch.nn.Linear(H, D_out),
            ...    )
            >>> optimizer = torch.optim.SGD(
            ...      model.parameters(),
            ...      lr=0.01,
            ...      momentum=0.9
            ...    )
            >>> sync_bn_model = bagua.torch_api.contrib.sync_batchnorm.SyncBatchNorm.convert_sync_batchnorm(model)
            >>> bagua_model = sync_bn_model.with_bagua([optimizer], GradientAllReduce())
        """
        module_output = module

        if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
            module_output = SyncBatchNorm(
                module.num_features,
                module.eps,
                module.momentum,
                module.affine,
                module.track_running_stats,
            )
            if module.affine:
                with torch.no_grad():
                    module_output.weight = module.weight
                    module_output.bias = module.bias
            module_output.running_mean = module.running_mean
            module_output.running_var = module.running_var
            module_output.num_batches_tracked = module.num_batches_tracked
            if hasattr(module, "qconfig"):
                module_output.qconfig = module.qconfig
        for name, child in module.named_children():
            module_output.add_module(name, cls.convert_sync_batchnorm(child))
        del module
        return module_output


class _SyncBatchNorm(Function):
    @staticmethod
    def forward(self, input, weight, bias, running_mean, running_var, eps, momentum):
        input = input.contiguous()

        size = input.numel() // input.size(1)
        count = torch.tensor([size])

        # calculate mean/invstd for input.
        mean, invstd = torch.batch_norm_stats(input, eps)
        count, mean, invstd = count.cuda(), mean.cuda(), invstd.cuda()

        nums_ranks = bagua.get_world_size()
        count_all = torch.tensor(
            [torch.empty_like(count).cpu().detach().numpy() for _ in range(nums_ranks)]
        ).cuda()
        mean_all = torch.tensor(
            [torch.empty_like(mean).cpu().detach().numpy() for _ in range(nums_ranks)]
        ).cuda()
        invstd_all = torch.tensor(
            [torch.empty_like(invstd).cpu().detach().numpy() for _ in range(nums_ranks)]
        ).cuda()

        allgather(count.unsqueeze(0), count_all)
        allgather(mean.unsqueeze(0), mean_all)
        allgather(invstd.unsqueeze(0), invstd_all)

        if _SYNC_BN_V3:
            counts_for_bngswc = count_all.view(-1).float().to(input.device)
        else:
            # backwards compatibility
            counts_for_bngswc = count_all.view(-1).tolist()

        # calculate global mean & invstd
        mean, invstd = torch.batch_norm_gather_stats_with_counts(
            input,
            mean_all,
            invstd_all,
            running_mean,
            running_var,
            momentum,
            eps,
            counts_for_bngswc,
        )

        self.save_for_backward(input, weight, mean, invstd, count_all)

        # apply element-wise normalization
        return torch.batch_norm_elemt(input, weight, bias, mean, invstd, eps)

    @staticmethod
    def backward(self, grad_output):
        grad_output = grad_output.contiguous()
        saved_input, weight, mean, invstd, count_all = self.saved_tensors
        need_input_grad, need_weight_grad, need_bias_grad = self.needs_input_grad[0:3]

        # calculate local stats as well as grad_weight / grad_bias
        sum_dy, sum_dy_xmu, grad_weight, grad_bias = torch.batch_norm_backward_reduce(
            grad_output,
            saved_input,
            mean,
            invstd,
            weight,
            need_input_grad,
            need_weight_grad,
            need_bias_grad,
        )

        if need_input_grad:
            # synchronizing stats used to calculate input gradient.
            allreduce(sum_dy, sum_dy)
            allreduce(sum_dy_xmu, sum_dy_xmu)

            if _SYNC_BN_V4:
                # from 1.9.0 on we need a count tensor on all devices
                # count_all is calculated as total count across all ranks in forward function
                count_all = count_all.to(dtype=torch.int, device=grad_output.device)
            elif _SYNC_BN_V2 or _SYNC_BN_V3:
                # before 1.9.0 we need the count as an integer to compute means values
                count = count_all.sum()
            else:
                # before 1.5.0, sum_dy was sum of means from every worker, so we just
                # need to divide it by number of workers
                count = bagua.get_world_size()

            # backward pass for gradient calculation
            # we are calling into a non-public undocumented function which broke moving to 1.9.0
            # https://github.com/pytorch/pytorch/issues/57900
            if _SYNC_BN_V4:
                # from 1.9.0 on, sums and count parameters expected
                grad_input = torch.batch_norm_backward_elemt(
                    grad_output,
                    saved_input,
                    mean,
                    invstd,
                    weight,
                    sum_dy,
                    sum_dy_xmu,
                    count_all,
                )
            else:
                # before 1.9.0, mean parameters expected, not sums and count
                grad_input = torch.batch_norm_backward_elemt(
                    grad_output,
                    saved_input,
                    mean,
                    invstd,
                    weight,
                    sum_dy / count,
                    sum_dy_xmu / count,
                )
        else:
            grad_input = None

        # synchronizing of grad_weight / grad_bias is not needed as distributed
        # training would handle all reduce.
        if weight is None or not need_weight_grad:
            grad_weight = None

        if weight is None or not need_bias_grad:
            grad_bias = None

        return grad_input, grad_weight, grad_bias, None, None, None, None, None, None



import ctypes
import os
import pathlib


def _preload_libraries():
    libnccl_path = os.path.join(
        pathlib.Path.home(), ".local", "share", "bagua", "nccl", "lib", "libnccl.so"
    )
    if os.path.exists(libnccl_path):
        ctypes.CDLL(libnccl_path)



#!/usr/bin/env python3

from bagua.torch_api.bucket import BaguaBucket
from bagua.torch_api.tensor import BaguaTensor
from bagua.torch_api.data_parallel.bagua_distributed import BaguaDistributedDataParallel
from bagua.torch_api.algorithms import Algorithm, AlgorithmImpl
from bagua.torch_api.communication import BaguaProcessGroup
from typing import List


class ByteGradAlgorithmImpl(AlgorithmImpl):
    def __init__(
        self,
        process_group: BaguaProcessGroup,
        hierarchical: bool = True,
        average: bool = True,
    ):
        """
        Implementation of the
        `ByteGrad <https://tutorials.baguasys.com/algorithms/bytegrad>`_
        algorithm.

        Args:
            process_group (BaguaProcessGroup): The process group to work on.
            hierarchical (bool): Enable hierarchical communication.
            average (bool): If ``True``, the gradients on each worker are averaged.
                Otherwise, they are summed.
        """
        super(ByteGradAlgorithmImpl, self).__init__(process_group)
        self.hierarchical = hierarchical
        self.average = average

    def tensors_to_buckets(
        self, tensors: List[List[BaguaTensor]], do_flatten: bool
    ) -> List[BaguaBucket]:
        bagua_buckets = []
        for idx, bucket in enumerate(tensors):
            bagua_bucket = BaguaBucket(
                bucket,
                flatten=do_flatten,
                name=str(idx),
                alignment=self.process_group.get_global_communicator().nranks(),
            )
            bagua_buckets.append(bagua_bucket)
        return bagua_buckets

    def init_operations(
        self,
        bagua_ddp: BaguaDistributedDataParallel,
        bucket: BaguaBucket,
    ):
        bucket.clear_ops()
        bucket.append_centralized_synchronous_op(
            hierarchical=self.hierarchical,
            average=self.average,
            scattergather=True,
            compression="MinMaxUInt8",
            group=self.process_group,
        )


class ByteGradAlgorithm(Algorithm):
    def __init__(self, hierarchical: bool = True, average: bool = True):
        """
        Create an instance of the
        `ByteGrad <https://tutorials.baguasys.com/algorithms/bytegrad>`_
        algorithm.

        Args:
            hierarchical (bool): Enable hierarchical communication.
            average (bool): If ``True``, the gradients on each worker are averaged.
                Otherwise, they are summed.
        """
        self.hierarchical = hierarchical
        self.average = average

    def reify(self, process_group: BaguaProcessGroup) -> ByteGradAlgorithmImpl:
        return ByteGradAlgorithmImpl(
            process_group,
            hierarchical=self.hierarchical,
            average=self.average,
        )



import os
import socket


def get_world_size() -> int:
    """
    Get the number of processes in the default process group.

    Returns:
        The world size of the default process group.
    """
    return int(os.environ.get("WORLD_SIZE", 1))


def get_rank() -> int:
    """
    Get the rank of the default process group.

    Rank is a unique identifier assigned to each process within the default
    process group. They are always consecutive integers ranging from 0 to
    ``world_size``.

    Returns:
        The rank of the default process group.
    """
    return int(os.environ.get("RANK", 0))


def get_local_rank() -> int:
    """
    Get the rank of current node.

    Local rank is a unique identifier assigned to each process within a node.
    They are always consecutive integers ranging from 0 to ``local_size``.

    Returns:
        The local rank of the node.
    """
    return int(os.environ.get("LOCAL_RANK", 0))


def get_local_size() -> int:
    """
    Get the number of processes in the node.

    Returns:
        The local size of the node.
    """
    return int(os.environ.get("LOCAL_WORLD_SIZE", 1))


def get_node_rank() -> int:
    """
    Get the rank among all nodes.

    Returns:
        The node rank of the node.
    """
    if _is_elastic_launched():
        return int(os.environ.get("GROUP_RANK", 0))
    else:
        return int(os.environ.get("NODE_RANK", 0))


def _is_elastic_launched():
    """Returns ``True`` if the current process was launched using the bagua.distributed.run command."""
    required_env_vars = {"RANK", "GROUP_RANK", "LOCAL_RANK", "LOCAL_WORLD_SIZE"}
    return required_env_vars.issubset(os.environ.keys())


def get_default_bucket_size() -> int:
    """Get default communication bucket byte size.

    Returns:
        The default bucket size.
    """
    return int(os.environ.get("BAGUA_DEFAULT_BUCKET_SIZE", 10 * 1024 ** 2))


def get_master_addr() -> str:
    return os.environ.get("MASTER_ADDR", "127.0.0.1")


def get_bagua_service_port() -> int:
    return int(os.environ.get("BAGUA_SERVICE_PORT", -1))


def is_report_metrics_switch_on() -> bool:
    """
    Whether bagua report switch is on or not.
    """
    return int(os.environ.get("BAGUA_REPORT_METRICS", 0)) == 1


# ** Autotune Environment Variable **


def get_autotune_level() -> int:
    """Get the autotune level.

    Returns:
        The autotune level.
    """
    return int(os.environ.get("BAGUA_AUTOTUNE", 0))


def get_autotune_max_samples() -> int:
    return int(os.environ.get("BAGUA_AUTOTUNE_MAX_SAMPLES", 60))


def get_autotune_sampling_confidence_time_s() -> float:
    return float(os.environ.get("BAGUA_AUTOTUNE_SAMPLING_CONFIDENCE_TIME_S", 5.0))


def get_autotune_warmup_time_s() -> float:
    return float(os.environ.get("BAGUA_AUTOTUNE_WARMUP_TIME_S", 30.0))


def get_is_output_autotune_log() -> bool:
    return bool(os.environ.get("BAGUA_IS_OUTPUT_AUTOTUNE_LOG", 0))


def get_autotune_server_wait_time() -> int:
    return int(os.environ.get("BAGUA_AUTOTUNE_SERVER_WAIT_TIME", 300))


def find_free_network_port() -> int:
    """Finds a free port on localhost."""
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port



# From PyTorch:
#
# Copyright (c) 2016-     Facebook, Inc            (Adam Paszke)
# Copyright (c) 2014-     Facebook, Inc            (Soumith Chintala)
# Copyright (c) 2011-2014 Idiap Research Institute (Ronan Collobert)
# Copyright (c) 2012-2014 Deepmind Technologies    (Koray Kavukcuoglu)
# Copyright (c) 2011-2012 NEC Laboratories America (Koray Kavukcuoglu)
# Copyright (c) 2011-2013 NYU                      (Clement Farabet)
# Copyright (c) 2006-2010 NEC Laboratories America (Ronan Collobert, Leon Bottou, Iain Melvin, Jason Weston)
# Copyright (c) 2006      Idiap Research Institute (Samy Bengio)
# Copyright (c) 2001-2004 Idiap Research Institute (Ronan Collobert, Samy Bengio, Johnny Mariethoz)
#
# From Caffe2:
#
# Copyright (c) 2016-present, Facebook Inc. All rights reserved.
#
# All contributions by Facebook:
# Copyright (c) 2016 Facebook Inc.
#
# All contributions by Google:
# Copyright (c) 2015 Google Inc.
# All rights reserved.
#
# All contributions by Yangqing Jia:
# Copyright (c) 2015 Yangqing Jia
# All rights reserved.
#
# All contributions by Kakao Brain:
# Copyright 2019-2020 Kakao Brain
#
# All contributions from Caffe:
# Copyright(c) 2013, 2014, 2015, the respective contributors
# All rights reserved.
#
# All other contributions:
# Copyright(c) 2015, 2016 the respective contributors
# All rights reserved.
#
# Caffe2 uses a copyright model similar to Caffe: each contributor holds
# copyright over their contributions to Caffe2. The project versioning records
# all such contribution and copyright details. If a contributor wants to further
# mark their specific copyright on a particular contribution, they should
# indicate their copyright solely in the commit message of the change when it is
# committed.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the names of Facebook, Deepmind Technologies, NYU, NEC Laboratories America
# and IDIAP Research Institute nor the names of its contributors may be
# used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# This file is copied from https://github.com/pytorch/pytorch/tree/v1.9.0/torch/testing/_internal/expecttest.py
import re
import unittest
import traceback
import os
import string
from typing import Tuple


# This file implements expect tests (also known as "golden" tests).
# Expect tests are a method of writing tests where instead of
# hard-coding the expected output of a test, you instead run the test to
# get the output, and the test framework automatically populates the
# expected output.  If the output of the test changes, you can rerun the
# test with EXPECTTEST_ACCEPT=1 environment variable to automatically
# update the expected output.
#
# Somewhat unusually, this file implements *inline* expect tests: that
# is to say, the expected output isn't save to an external file, it is
# saved directly in the Python file (and we modify your Python the file
# when updating the expect test.)
#
# The general recipe for how to use this is as follows:
#
#   1. Write your test and use assertExpectedInline() instead of
#      a normal assertEqual.  Leave the expected argument blank
#      with an empty string:
#
#       self.assertExpectedInline(some_func(), "")
#
#   2. Run your test.  It should fail, and you get an error message
#      about accepting the output with EXPECTTEST_ACCEPT=1
#
#   3. Rerun the test with EXPECTTEST_ACCEPT=1.  Now the previously
#      blank string literal will now contain the expected value of
#      the test.
#
#       self.assertExpectedInline(some_func(), "my_value")
#
# Some tips and tricks:
#
#   - Often, you will want to expect test on a multiline string.  This
#     framework understands triple-quoted strings, so you can just
#     write """my_value""" and it will turn into triple-quoted
#     strings.
#
#   - Take some time thinking about how exactly you want to design
#     the output format of the expect test.  It is often profitable
#     to design an output representation specifically for expect tests.
#


ACCEPT = os.getenv("EXPECTTEST_ACCEPT")


def nth_line(src, lineno):
    """
    Compute the starting index of the n-th line (where n is 1-indexed)

    >>> nth_line("aaa\\nbb\\nc", 2)
    4
    """
    assert lineno >= 1
    pos = 0
    for _ in range(lineno - 1):
        pos = src.find("\n", pos) + 1
    return pos


def nth_eol(src, lineno):
    """
    Compute the ending index of the n-th line (before the newline,
    where n is 1-indexed)

    >>> nth_eol("aaa\\nbb\\nc", 2)
    6
    """
    assert lineno >= 1
    pos = -1
    for _ in range(lineno):
        pos = src.find("\n", pos + 1)
        if pos == -1:
            return len(src)
    return pos


def normalize_nl(t):
    return t.replace("\r\n", "\n").replace("\r", "\n")


def escape_trailing_quote(s, quote):
    if s and s[-1] == quote:
        return s[:-1] + "\\" + quote
    else:
        return s


class EditHistory(object):
    def __init__(self):
        self.state = {}

    def adjust_lineno(self, fn, lineno):
        if fn not in self.state:
            return lineno
        for edit_loc, edit_diff in self.state[fn]:
            if lineno > edit_loc:
                lineno += edit_diff
        return lineno

    def seen_file(self, fn):
        return fn in self.state

    def record_edit(self, fn, lineno, delta):
        self.state.setdefault(fn, []).append((lineno, delta))


EDIT_HISTORY = EditHistory()


def ok_for_raw_triple_quoted_string(s, quote):
    """
    Is this string representable inside a raw triple-quoted string?
    Due to the fact that backslashes are always treated literally,
    some strings are not representable.

    >>> ok_for_raw_triple_quoted_string("blah", quote="'")
    True
    >>> ok_for_raw_triple_quoted_string("'", quote="'")
    False
    >>> ok_for_raw_triple_quoted_string("a ''' b", quote="'")
    False
    """
    return quote * 3 not in s and (not s or s[-1] not in [quote, "\\"])


# This operates on the REVERSED string (that's why suffix is first)
RE_EXPECT = re.compile(
    r"^(?P<suffix>[^\n]*?)"
    r"(?P<quote>'''|"
    r'""")'
    r"(?P<body>.*?)"
    r"(?P=quote)"
    r"(?P<raw>r?)",
    re.DOTALL,
)


def replace_string_literal(src: str, lineno: int, new_string: str) -> Tuple[str, int]:
    r"""
    Replace a triple quoted string literal with new contents.
    Only handles printable ASCII correctly at the moment.  This
    will preserve the quote style of the original string, and
    makes a best effort to preserve raw-ness (unless it is impossible
    to do so.)

    Returns a tuple of the replaced string, as well as a delta of
    number of lines added/removed.

    >>> replace_string_literal("'''arf'''", 1, "barf")
    ("'''barf'''", 0)
    >>> r = replace_string_literal("  moo = '''arf'''", 1, "'a'\n\\b\n")
    >>> print(r[0])
      moo = '''\
    'a'
    \\b
    '''
    >>> r[1]
    3
    >>> replace_string_literal("  moo = '''\\\narf'''", 2, "'a'\n\\b\n")[1]
    2
    >>> print(replace_string_literal("    f('''\"\"\"''')", 1, "a ''' b")[0])
        f('''a \'\'\' b''')
    """
    # Haven't implemented correct escaping for non-printable characters
    assert all(c in string.printable for c in new_string)
    i = nth_eol(src, lineno)
    new_string = normalize_nl(new_string)

    delta = [new_string.count("\n")]
    if delta[0] > 0:
        delta[0] += 1  # handle the extra \\\n

    def replace(m):
        s = new_string
        raw = m.group("raw") == "r"
        if not raw or not ok_for_raw_triple_quoted_string(s, quote=m.group("quote")[0]):
            raw = False
            s = s.replace("\\", "\\\\")
            if m.group("quote") == "'''":
                s = escape_trailing_quote(s, "'").replace("'''", r"\'\'\'")
            else:
                s = escape_trailing_quote(s, '"').replace('"""', r"\"\"\"")

        new_body = "\\\n" + s if "\n" in s and not raw else s
        delta[0] -= m.group("body").count("\n")

        return "".join(
            [
                m.group("suffix"),
                m.group("quote"),
                new_body[::-1],
                m.group("quote"),
                "r" if raw else "",
            ]
        )

    # Having to do this in reverse is very irritating, but it's the
    # only way to make the non-greedy matches work correctly.
    return (RE_EXPECT.sub(replace, src[:i][::-1], count=1)[::-1] + src[i:], delta[0])


class TestCase(unittest.TestCase):
    longMessage = True

    def assertExpectedInline(self, actual, expect, skip=0):
        """
        Assert that actual is equal to expect.  The expect argument
        MUST be a string literal (triple-quoted strings OK), and will
        get updated directly in source when you run the test suite
        with EXPECTTEST_ACCEPT=1.

        If you want to write a helper function that makes use of
        assertExpectedInline (e.g., expect is not a string literal),
        set the skip argument to how many function calls we should
        skip to find the string literal to update.
        """
        if ACCEPT:
            if actual != expect:
                # current frame and parent frame, plus any requested skip
                tb = traceback.extract_stack(limit=2 + skip)
                fn, lineno, _, _ = tb[0]
                print(
                    "Accepting new output for {} at {}:{}".format(self.id(), fn, lineno)
                )
                with open(fn, "r+") as f:
                    old = f.read()

                    # compute the change in lineno
                    lineno = EDIT_HISTORY.adjust_lineno(fn, lineno)
                    new, delta = replace_string_literal(old, lineno, actual)

                    assert (
                        old != new
                    ), f"Failed to substitute string at {fn}:{lineno}; did you use triple quotes?"

                    # Only write the backup file the first time we hit the
                    # file
                    if not EDIT_HISTORY.seen_file(fn):
                        with open(fn + ".bak", "w") as f_bak:
                            f_bak.write(old)
                    f.seek(0)
                    f.truncate(0)

                    f.write(new)

                EDIT_HISTORY.record_edit(fn, lineno, delta)
        else:
            help_text = (
                "To accept the new output, re-run test with "
                "envvar EXPECTTEST_ACCEPT=1 (we recommend "
                "staging/committing your changes before doing this)"
            )
            self.assertMultiLineEqualMaybeCppStack(expect, actual, msg=help_text)

    def assertExpectedRaisesInline(self, exc_type, callable, expect, *args, **kwargs):
        """
        Like assertExpectedInline, but tests the str() representation of
        the raised exception from callable.  The raised exeption must
        be exc_type.
        """
        try:
            callable(*args, **kwargs)
        except exc_type as e:
            self.assertExpectedInline(str(e), expect, skip=1)
            return
        # Don't put this in the try block; the AssertionError will catch it
        self.fail(msg="Did not raise when expected to")

    def assertMultiLineEqualMaybeCppStack(self, expect, actual, *args, **kwargs):
        self.assertGreaterEqual(len(actual), len(expect), *args, **kwargs)
        if hasattr(self, "assertMultiLineEqual"):
            self.assertMultiLineEqual(expect, actual[: len(expect)], *args, **kwargs)
        else:
            self.assertEqual(expect, actual[: len(expect)], *args, **kwargs)
        if len(actual) > len(expect):
            cpp_stacktrace_header = "\nException raised from"
            end_header = len(expect) + len(cpp_stacktrace_header)
            self.assertEqual(actual[len(expect) : end_header], cpp_stacktrace_header)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from .fuse.optimizer import fuse_optimizer  # noqa: F401
from .load_balancing_data_loader import (  # noqa: F401
    LoadBalancingDistributedSampler,
    LoadBalancingDistributedBatchSampler,
)
from .cache_loader import CacheLoader  # noqa: F401
from .cached_dataset import CachedDataset  # noqa: F401



#!/usr/bin/env python

"""
Multi-node launch script. execute bagua.distributed.launch on the specified host list.
The following is an example of running bagua program with two nodes:
```bash
# Started by baguarun
Launch bagua on two nodes *(192.168.1.1, 192.168.1.2)*
    >>> baguarun --host_list 192.168.1.1,192.168.1.2 --ssh_port ${COMM_SSH_PORT}
                --nproc_per_node=NUM_GPUS_YOU_HAVE --master_port=1234
                YOUR_TRAINING_SCRIPT.py (--arg1 --arg2 --arg3 and all other arguments of your training script)
# The same operation do with bagua.distributed.launch
Node 1: *(IP: 192.168.1.1, and has a free port: 1234)*
::
    >>> python -m bagua.distributed.launch --nproc_per_node=NUM_GPUS_YOU_HAVE
               --nnodes=2 --node_rank=0 --master_addr="192.168.1.1"
               --master_port=1234 YOUR_TRAINING_SCRIPT.py (--arg1 --arg2 --arg3
               and all other arguments of your training script)
Node 2 *(IP: 192.168.1.2):
::
    >>> python -m bagua.distributed.launch --nproc_per_node=NUM_GPUS_YOU_HAVE
               --nnodes=2 --node_rank=1 --master_addr="192.168.1.1"
               --master_port=1234 YOUR_TRAINING_SCRIPT.py (--arg1 --arg2 --arg3
               and all other arguments of your training script)
```
"""

import argparse
import os
import gevent
from pssh.clients import ParallelSSHClient
from pssh.config import HostConfig
from pssh.utils import enable_host_logger


def pssh_bagua_launch(
    args,
    script_cmd: str,
    env: dict = {},
):
    host_list = [host for (host, _) in args.host_list]
    nproc_per_node = args.nproc_per_node
    assert len(host_list) != 0, "Invalid host_list={}".format(host_list)

    if "PATH" not in env:
        env["PATH"] = os.environ["PATH"]
    if "LD_LIBRARY_PATH" not in env:
        env["LD_LIBRARY_PATH"] = os.environ["LD_LIBRARY_PATH"]

    pretreat_cmd = [
        "shopt -s huponexit; cd {};".format(os.getcwd()),
    ]
    for k, v in env.items():
        pretreat_cmd.append(
            "export {key}={value} &&".format(
                key=k,
                value=v,
            )
        )

    bypass_args = []
    if args.master_port:
        bypass_args.append("--master_port={}".format(args.master_port))
    if args.bagua_service_port:
        bypass_args.append(
            "--bagua_service_port={}".format(args.bagua_service_port)
        )  # noqa: E501
    if args.no_python:
        bypass_args.append("--no_python")
    if args.enable_bagua_net:
        bypass_args.append("--enable_bagua_net")

    host_config = [HostConfig(port=port) for (_, port) in args.host_list]

    master_addr = host_list[0]
    host_args = []
    for i, _ in enumerate(host_list):
        host_args.append(
            {
                "cmd": " ".join(
                    pretreat_cmd
                    + [
                        "python -m bagua.distributed.launch",
                        "--nproc_per_node={}".format(nproc_per_node),
                        "--nnodes={} --node_rank={}".format(len(host_list), i),
                        '--master_addr="{}"'.format(master_addr),
                    ]
                    + bypass_args
                    + [script_cmd]
                ),
            }
        )

    enable_host_logger()
    client = ParallelSSHClient(
        host_list, host_config=host_config, pool_size=10 + len(host_list)
    )
    output = client.run_command(
        "%(cmd)s",
        host_args=host_args,
        shell="bash -xc",
        use_pty=True,  # The key configuration of process safe exit
    )

    def do_output(host_out):
        for line in host_out.stdout:
            pass

    cmds = [client.pool.spawn(do_output, host_out) for host_out in output]
    gevent.joinall(cmds)

    client.join()


def parse_args():
    parser = argparse.ArgumentParser(description="Distributed bagua launch")
    parser.add_argument(
        "--host_list",
        default=None,
        help="Target node list, baguarun will access each node via ssh."
        "`--host_list` has two forms. The first, comma-separated domain name "
        "or IP, such as `--host_list 192.168.1.1,192.168.1.2`. In this case. the ssh port is "
        "specified by the `--ssh_port` argument. The second, domain name or IP with ssh "
        "port specified inline, such as `--host_list 192.168.1.1:22,192.168.1.2:8022`."
        "\nIf `--host_list` is None, baguarun will use the "
        "environment variable BAGUA_NODE_DOMAIN_NAMES.",
    )
    parser.add_argument(
        "--ssh_port",
        type=int,
        default=None,
        help="ssh port to use. you need to set it only when the ssh "
        "port of all nodes are the same. When "
        "ssh_port is not set, baguaru will use the "
        "environment variable BAGUA_SSH_PORT.",
    )
    parser.add_argument("--master_port", type=int, default=None)
    parser.add_argument("--nproc_per_node", type=int, required=True)
    parser.add_argument(
        "--no_python",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--bagua_service_port",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--enable_bagua_net",
        action="store_true",
        default=False,
        help="Enable Bagua-Net optimization for better "
        "communication performance. See https://github.com/BaguaSys/bagua/tree/master/rust/bagua-net"
        " for more details.",
    )
    parser.add_argument(
        "-x",
        type=str,
        action="append",
        help="Environment variables that need to be passed in pssh.",
    )

    # positional
    parser.add_argument(
        "training_script",
        type=str,
        help="The full path to the single GPU training "
        "program/script to be launched in parallel, "
        "followed by all the arguments for the "
        "training script",
    )

    # rest from the training program
    parser.add_argument("training_script_args", nargs=argparse.REMAINDER)

    args = parser.parse_args()

    if args.host_list is None:
        args.host_list = os.environ.get("BAGUA_NODE_DOMAIN_NAMES", "")
    if args.ssh_port is None and "BAGUA_SSH_PORT" in os.environ:
        args.ssh_port = int(os.environ["BAGUA_SSH_PORT"])

    assert args.host_list, "`--host_list` or $BAGUA_NODE_DOMAIN_NAMES must be set"
    host_list = []
    for host_str in args.host_list.split(","):
        host_or_pair = host_str.split(":")
        if len(host_or_pair) == 1:
            assert args.ssh_port, "`--host_list` or $BAGUA_SSH_PORT must be set"
            host_list.append((host_or_pair[0], args.ssh_port))
        elif len(host_or_pair) == 2:
            (host, port) = host_or_pair
            host_list.append((host, int(port)))
        else:
            raise Exception("Invalid host={}".format(host_str))

    args.host_list = host_list

    args.set_env = {}
    if args.x is not None:
        for x_word in args.x:
            if "=" in x_word:
                (k, v) = x_word.split("=", 1)
                args.set_env[k] = v
            else:
                k = x_word
                args.set_env[k] = os.environ.get(k, "")

    return args


def main():
    args = parse_args()
    pssh_bagua_launch(
        args,
        script_cmd=" ".join(
            [
                args.training_script,
            ]
            + args.training_script_args
        ),
        env=args.set_env,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
The Bagua communication library PyTorch interface.
"""
from distutils.errors import (
    DistutilsPlatformError,
)

try:
    import torch
except ImportError:
    print("import torch failed, is it installed?")

version = torch.__version__
if version is None:
    raise DistutilsPlatformError(
        "Unable to determine PyTorch version from the version string '%s'"
        % torch.__version__
    )
elif version < "1.6.0":
    raise Exception(
        "Bagua need pytorch version >= 1.6.0, while current version is {}.".format(
            version
        )
    )

from .communication import (  # noqa: F401
    get_backend,
    init_process_group,
    send,
    recv,
    broadcast,
    reduce,
    reduce_inplace,
    gather,
    gather_inplace,
    scatter,
    scatter_inplace,
    allreduce,
    allreduce_inplace,
    allgather,
    allgather_inplace,
    alltoall,
    alltoall_inplace,
    alltoall_v,
    alltoall_v_inplace,
    reduce_scatter,
    reduce_scatter_inplace,
    ReduceOp,
)
from .distributed import BaguaModule  # noqa: F401
from .tensor import BaguaTensor  # noqa: F401
from .env import (  # noqa: F401
    get_rank,
    get_world_size,
    get_local_rank,
    get_local_size,
)
from . import contrib  # noqa: F401
from . import communication  # noqa: F401
from . import algorithms  # noqa: F401
from . import checkpoint  # noqa: E402,F401
from . import data_parallel  # noqa: E402,F401
from .model_parallel import moe  # noqa: E402,F401
